import re
import unittest
from binascii import hexlify

from Cryptodome.Util.py3compat import bord

from Cryptodome.Hash import SHA256
from Cryptodome.PublicKey import ECC
from Cryptodome.SelfTest.st_common import list_test_cases
from Cryptodome.SelfTest.loader import load_test_vectors, load_test_vectors_wycheproof

from Cryptodome.Protocol.DH import key_agreement


class FIPS_ECDH_Tests_KAT(unittest.TestCase):
    pass


test_vectors_verify = load_test_vectors(("Protocol", ),
                                        "KAS_ECC_CDH_PrimitiveTest.txt",
                                        "ECC CDH Primitive (SP800-56A Section 5.7.1.2)",
                                        {
                                        'qcavsx': lambda x: int(x, 16),
                                        'qcavsy': lambda x: int(x, 16),
                                        'diut': lambda x: int(x, 16),
                                        'qiutx': lambda x: int(x, 16),
                                        'qiuty': lambda x: int(x, 16),
                                        }) or []

for idx, tv in enumerate(test_vectors_verify):

    # Stand-alone header with curve name
    if isinstance(tv, str):
        res = re.match(r"\[([A-Za-z0-9-]+)\]", tv)
        assert res
        curve_name = res.group(1)
        continue

    public_key = ECC.construct(curve=curve_name,
                               point_x=tv.qcavsx,
                               point_y=tv.qcavsy)

    private_key = ECC.construct(curve=curve_name,
                                d=tv.diut)

    exp_response = tv.ziut

    def ecdh_test(self,
                  public_key=public_key,
                  private_key=private_key,
                  exp_response=exp_response):
        z = key_agreement(
                static_pub=public_key,
                static_priv=private_key,
                kdf=lambda x: x)
        self.assertEqual(z, exp_response)

    def ecdh_test_rev(self,
                      public_key=public_key,
                      private_key=private_key,
                      exp_response=exp_response):
        z = key_agreement(
                static_pub=public_key,
                static_priv=private_key,
                kdf=lambda x: x)
        self.assertEqual(z, exp_response)

    setattr(FIPS_ECDH_Tests_KAT, "test_verify_positive_%d" % idx, ecdh_test)
    if idx == 1:
        setattr(FIPS_ECDH_Tests_KAT, "test_verify_positive_rev_%d" % idx, ecdh_test_rev)


class TestVectorsECDHWycheproof(unittest.TestCase):

    desc = "Wycheproof ECDH tests"

    def add_tests(self, filename):

        def curve(g):
            return g['curve']

        def private(u):
            return int(u['private'], 16)

        result = load_test_vectors_wycheproof(("Protocol", "wycheproof"),
                                              filename,
                                              "Wycheproof ECDH (%s)"
                                              % filename,
                                              group_tag={'curve': curve},
                                              unit_tag={'private': private},
                                              )
        self.tv += result

    def setUp(self):
        self.tv = []
        self.desc = None

        self.add_tests("ecdh_secp224r1_ecpoint_test.json")
        self.add_tests("ecdh_secp256r1_ecpoint_test.json")
        self.add_tests("ecdh_secp384r1_ecpoint_test.json")
        self.add_tests("ecdh_secp521r1_ecpoint_test.json")

        self.add_tests("ecdh_secp224r1_test.json")
        self.add_tests("ecdh_secp256r1_test.json")
        self.add_tests("ecdh_secp384r1_test.json")
        self.add_tests("ecdh_secp521r1_test.json")

    def shortDescription(self):
        return self.desc

    def test_verify(self, tv):

        if len(tv.public) == 0:
            return

        try:
            if bord(tv.public[0]) == 4:     # SEC1
                public_key = ECC.import_key(tv.public, curve_name=tv.curve)
            else:
                public_key = ECC.import_key(tv.public)
        except ValueError:
            assert tv.warning or not tv.valid
            return

        private_key = ECC.construct(curve=tv.curve, d=tv.private)

        try:
            z = key_agreement(static_pub=public_key,
                              static_priv=private_key,
                              kdf=lambda x: x)
        except ValueError:
            assert not tv.valid
        except TypeError as e:
            assert not tv.valid
            assert "incompatible curve" in str(e)
        else:
            self.assertEqual(z, tv.shared)
            assert tv.valid

    def runTest(self):
        for tv in self.tv:
            self.desc = "Wycheproof ECDH Verify Test #%d (%s, %s)" % (tv.id, tv.comment, tv.filename)
            self.test_verify(tv)


class ECDH_Tests(unittest.TestCase):

    static_priv = ECC.import_key('-----BEGIN PRIVATE KEY-----\nMIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQg9VHFVKh2a1aVFifH\n+BiyNaRa2kttEg3165Ye/dJxJ7KhRANCAARImIEXro5ZOcyWU2mq/+d79FEZXtTA\nbKkz1aICQXihQdCMzRNbeNtC9LFLzhu1slRKJ2xsDAlw9r6w6vwtkRzr\n-----END PRIVATE KEY-----')
    static_pub = ECC.import_key('-----BEGIN PRIVATE KEY-----\nMIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQgHhmv8zmZ+Nw8fsZd\ns8tlZflyfw2NE1CRS9DWr3Y3O46hRANCAAS3hZVUCbk+uk3w4S/YOraEVGG+WYpk\nNO/vrwzufUUks2GV2OnBQESe0EBk4Jq8gn4ij8Lvs3rZX2yT+XfeATYd\n-----END PRIVATE KEY-----').public_key()

    eph_priv = ECC.import_key('-----BEGIN PRIVATE KEY-----\nMIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQgGPdJmFFFKzLPspIr\nE1T2cEjeIf4ajS9CpneP0e2b3AyhRANCAAQBexAA5BYDcXHs2KOksTYUsst4HhPt\nkp0zkgI2virc3OGJFNGPaCCPfFCQJHwLRaEpiq3SoQlgoBwSc8ZPsl3y\n-----END PRIVATE KEY-----')

    eph_pub = ECC.import_key('-----BEGIN PRIVATE KEY-----\nMIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQghaVZXElSEGEojFKF\nOU0JCpxWUWHvWQUR81gwWrOp76ShRANCAATi1Ib2K+YR3AckD8wxypWef7pw5PRw\ntBaB3RDPyE7IjHZC6yu1DbcXoCdtaw+F5DM+4zpl59n5ZaIy/Yl1BdIy\n-----END PRIVATE KEY-----')

    def test_1(self):
        # C(0, 2s)
        kdf = lambda x: SHA256.new(x).digest()
        z = key_agreement(
                kdf=kdf,
                static_pub=self.static_pub,
                static_priv=self.static_priv)
        self.assertEqual(hexlify(z),
                        b"3960a1101d1193cbaffef4cc7202ebff783c22c6d2e0d5d530ffc66dc197ea9c")

    def test_2(self):
        # C(2e, 2s)
        kdf = lambda x: SHA256.new(x).digest()
        z = key_agreement(
                kdf=kdf,
                static_pub=self.static_pub,
                static_priv=self.static_priv,
                eph_pub=self.eph_pub,
                eph_priv=self.eph_priv)
        self.assertEqual(hexlify(z),
                         b"7447b733d40c8fab2c633b3dc61e4a8c742f3a6af7e16fb0cc486f5bdb5d6ba2")

    def test_3(self):
        # C(1e, 2s)
        kdf = lambda x: SHA256.new(x).digest()
        z = key_agreement(
                kdf=kdf,
                static_pub=self.static_pub,
                static_priv=self.static_priv,
                eph_priv=self.eph_priv)
        self.assertEqual(hexlify(z),
                         b"9e977ae45f33bf67f285d064d83e6632bcafe3a7d33fe571233bab4794ace759")

    def test_4(self):
        # C(1e, 2s)
        kdf = lambda x: SHA256.new(x).digest()
        z = key_agreement(
                kdf=kdf,
                static_pub=self.static_pub,
                static_priv=self.static_priv,
                eph_pub=self.eph_pub)
        self.assertEqual(hexlify(z),
                         b"c9532df6aa7e9dbe5fe85da31ee25ff19c179c88691ec4b8328cc2036dcdadf2")

    def test_5(self):
        # C(2e, 1s) is not supported
        kdf = lambda x: SHA256.new(x).digest()
        self.assertRaises(ValueError,
                key_agreement,
                kdf=kdf,
                static_priv=self.static_priv,
                eph_pub=self.eph_pub,
                eph_priv=self.eph_priv)

    def test_6(self):
        # C(2e, 1s) is not supported
        kdf = lambda x: SHA256.new(x).digest()
        self.assertRaises(ValueError,
                key_agreement,
                kdf=kdf,
                static_pub=self.static_pub,
                eph_pub=self.eph_pub,
                eph_priv=self.eph_priv)

    def test_7(self):
        # C(2e, 0)
        kdf = lambda x: SHA256.new(x).digest()
        z = key_agreement(
                kdf=kdf,
                eph_pub=self.eph_pub,
                eph_priv=self.eph_priv)
        self.assertEqual(hexlify(z),
                         b"feb257ebe063078b1391aac07913283d7b642ad7df61b46dfc9cd6f420bb896a")

    def test_8(self):
        # C(1e, 1s)
        kdf = lambda x: SHA256.new(x).digest()
        z = key_agreement(
                kdf=kdf,
                static_priv=self.static_priv,
                eph_pub=self.eph_pub)
        self.assertEqual(hexlify(z),
                         b"ee4dc995117476ed57fd17ff0ed44e9f0466d46b929443bc0db9380317583b04")

    def test_9(self):
        # C(1e, 1s)
        kdf = lambda x: SHA256.new(x).digest()
        z = key_agreement(
                kdf=kdf,
                static_pub=self.static_pub,
                eph_priv=self.eph_priv)
        self.assertEqual(hexlify(z),
                         b"2351cc2014f7c40468fa072b5d30f706eeaeef7507311cd8e59bab3b43f03c51")

    def test_10(self):
        # No private (local) keys
        kdf = lambda x: SHA256.new(x).digest()
        self.assertRaises(ValueError,
                key_agreement,
                kdf=kdf,
                static_pub=self.static_pub,
                eph_pub=self.eph_pub)

    def test_11(self):
        # No public (peer) keys
        kdf = lambda x: SHA256.new(x).digest()
        self.assertRaises(ValueError,
                key_agreement,
                kdf=kdf,
                static_priv=self.static_priv,
                eph_priv=self.eph_priv)

    def test_12(self):
        # failure if kdf is missing
        self.assertRaises(ValueError,
                key_agreement,
                static_pub=self.static_pub,
                static_priv=self.static_priv)


def get_tests(config={}):

    tests = []
    tests += list_test_cases(FIPS_ECDH_Tests_KAT)
    tests += [TestVectorsECDHWycheproof()]
    tests += list_test_cases(ECDH_Tests)

    slow_tests = config.get('slow_tests')
    if slow_tests:
        pass

    return tests


if __name__ == '__main__':
    def suite():
        return unittest.TestSuite(get_tests())
    unittest.main(defaultTest='suite')



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import os

# Import from pygit2
from .ffi import ffi


def maybe_string(ptr):
    if not ptr:
        return None

    return ffi.string(ptr).decode('utf8')


def to_bytes(s, encoding='utf-8', errors='strict'):
    if s == ffi.NULL or s is None:
        return ffi.NULL

    if hasattr(s, '__fspath__'):
        s = os.fspath(s)

    if isinstance(s, bytes):
        return s

    return s.encode(encoding, errors)


def to_str(s):
    if hasattr(s, '__fspath__'):
        s = os.fspath(s)

    if type(s) is str:
        return s

    if type(s) is bytes:
        return s.decode()

    raise TypeError('unexpected type "%s"' % repr(s))


def strarray_to_strings(arr):
    l = [None] * arr.count
    for i in range(arr.count):
        l[i] = ffi.string(arr.strings[i]).decode('utf-8')

    return l


class StrArray:
    """A git_strarray wrapper

    Use this in order to get a git_strarray* to pass to libgit2 out of a
    list of strings. This has a context manager, which you should use, e.g.

        with StrArray(list_of_strings) as arr:
            C.git_function_that_takes_strarray(arr)
    """

    def __init__(self, l):
        # Allow passing in None as lg2 typically considers them the same as empty
        if l is None:
            self.array = ffi.NULL
            return

        if not isinstance(l, (list, tuple)):
            raise TypeError("Value must be a list")

        strings = [None] * len(l)
        for i in range(len(l)):
            li = l[i]
            if not isinstance(li, str) and not hasattr(li, '__fspath__'):
                raise TypeError("Value must be a string or PathLike object")

            strings[i] = ffi.new('char []', to_bytes(li))

        self._arr = ffi.new('char *[]', strings)
        self._strings = strings
        self.array = ffi.new('git_strarray *', [self._arr, len(strings)])

    def __enter__(self):
        return self.array

    def __exit__(self, type, value, traceback):
        pass


class GenericIterator:
    """Helper to easily implement an iterator.

    The constructor gets a container which must implement __len__ and
    __getitem__
    """

    def __init__(self, container):
        self.container = container
        self.length = len(container)
        self.idx = 0

    def next(self):
        return self.__next__()

    def __next__(self):
        idx = self.idx
        if idx >= self.length:
            raise StopIteration

        self.idx += 1
        return self.container[idx]



import errno
import time
import traceback
from socket import AF_INET, AF_INET6
from socket import error as socket_error
from socket import inet_ntop, inet_pton

from pyroute2 import config
from pyroute2.common import Dotkeys, View, basestring, dqn2int
from pyroute2.config import AF_BRIDGE
from pyroute2.ipdb.exceptions import (
    CommitException,
    CreateException,
    PartialCommitException,
)
from pyroute2.ipdb.linkedset import LinkedSet
from pyroute2.ipdb.transactional import (
    SYNC_TIMEOUT,
    Transactional,
    with_transaction,
)
from pyroute2.netlink import rtnl
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.rtnl.ifinfmsg import IFF_MASK, ifinfmsg

supported_kinds = (
    'bridge',
    'bond',
    'tuntap',
    'vxlan',
    'gre',
    'gretap',
    'ip6gre',
    'ip6gretap',
    'macvlan',
    'macvtap',
    'ipvlan',
    'vrf',
    'vti',
)

groups = (
    rtnl.RTMGRP_LINK
    | rtnl.RTMGRP_NEIGH
    | rtnl.RTMGRP_IPV4_IFADDR
    | rtnl.RTMGRP_IPV6_IFADDR
)


def _get_data_fields():
    global supported_kinds

    ret = []
    for data in supported_kinds:
        msg = ifinfmsg.ifinfo.data_map.get(data)
        if msg is not None:
            if getattr(msg, 'prefix', None) is not None:
                ret += [msg.nla2name(i[0]) for i in msg.nla_map]
            else:
                ret += [ifinfmsg.nla2name(i[0]) for i in msg.nla_map]
    return ret


def _br_time_check(x, y):
    return abs(x - y) < 5


class Interface(Transactional):
    '''
    Objects of this class represent network interface and
    all related objects:
    * addresses
    * (todo) neighbours
    * (todo) routes

    Interfaces provide transactional model and can act as
    context managers. Any attribute change implicitly
    starts a transaction. The transaction can be managed
    with three methods:
    * review() -- review changes
    * rollback() -- drop all the changes
    * commit() -- try to apply changes

    If anything will go wrong during transaction commit,
    it will be rolled back authomatically and an
    exception will be raised. Failed transaction review
    will be attached to the exception.
    '''

    _fields_cmp = {
        'flags': lambda x, y: x & y & IFF_MASK == y & IFF_MASK,
        'br_hello_time': _br_time_check,
        'br_max_age': _br_time_check,
        'br_ageing_time': _br_time_check,
        'br_forward_delay': _br_time_check,
        'br_mcast_membership_intvl': _br_time_check,
        'br_mcast_querier_intvl': _br_time_check,
        'br_mcast_query_intvl': _br_time_check,
        'br_mcast_query_response_intvl': _br_time_check,
        'br_mcast_startup_query_intvl': _br_time_check,
    }
    _virtual_fields = [
        'ipdb_scope',
        'ipdb_priority',
        'vlans',
        'ipaddr',
        'ports',
        'vlan_flags',
        'net_ns_fd',
        'net_ns_pid',
    ]
    _fields = [ifinfmsg.nla2name(i[0]) for i in ifinfmsg.nla_map]
    for name in ('bridge_slave_data',):
        data = getattr(ifinfmsg.ifinfo, name)
        _fields.extend([ifinfmsg.nla2name(i[0]) for i in data.nla_map])
    _fields.append('index')
    _fields.append('flags')
    _fields.append('mask')
    _fields.append('change')
    _fields.append('kind')
    _fields.append('peer')
    _fields.append('vlan_id')
    _fields.append('vlan_protocol')
    _fields.append('bond_mode')
    _fields.extend(_get_data_fields())
    _fields.extend(_virtual_fields)

    def __init__(self, ipdb, mode=None, parent=None, uid=None):
        '''
        Parameters:
        * ipdb -- ipdb() reference
        * mode -- transaction mode
        '''
        Transactional.__init__(self, ipdb, mode)
        self.cleanup = (
            'header',
            'linkinfo',
            'protinfo',
            'af_spec',
            'attrs',
            'event',
            'map',
            'stats',
            'stats64',
            'change',
            '__align',
        )
        self.ingress = None
        self.egress = None
        self.nlmsg = None
        self.errors = []
        self.partial = False
        self._exception = None
        self._deferred_link = None
        self._tb = None
        self._linked_sets.add('ipaddr')
        self._linked_sets.add('ports')
        self._linked_sets.add('vlans')
        self._freeze = None
        self._delay_add_port = set()
        self._delay_del_port = set()
        # 8<-----------------------------------
        # local setup: direct state is required
        with self._direct_state:
            for i in ('change', 'mask'):
                del self[i]
            self['ipaddr'] = self.ipdb._ipaddr_set()
            self['ports'] = LinkedSet()
            self['vlans'] = LinkedSet()
            self['ipdb_priority'] = 0
        # 8<-----------------------------------

    def __hash__(self):
        return self['index']

    @property
    def if_master(self):
        '''
        [property] Link to the parent interface -- if it exists
        '''
        return self.get('master', None)

    def detach(self):
        self.ipdb.interfaces._detach(self['ifname'], self['index'], self.nlmsg)
        return self

    def freeze(self):
        if self._freeze is not None:
            raise RuntimeError("the interface is frozen already")

        dump = self.pick()

        def cb(ipdb, msg, action):
            if msg.get('index', -1) == dump['index']:
                try:
                    # important: that's a rollback, so do not
                    # try to revert changes in the case of failure
                    self.commit(
                        transaction=dump, commit_phase=2, commit_mask=2
                    )
                except Exception:
                    pass

        self._freeze = self.ipdb.register_callback(cb)
        return self

    def unfreeze(self):
        self.ipdb.unregister_callback(self._freeze)
        self._freeze = None
        return self

    def load(self, data):
        '''
        Load the data from a dictionary to an existing
        transaction. Requires `commit()` call, or must be
        called from within a `with` statement.

        Sample::

            data = json.loads(...)
            with ipdb.interfaces['dummy1'] as i:
                i.load(data)

        Sample, mode `explicit::

            data = json.loads(...)
            i = ipdb.interfaces['dummy1']
            i.begin()
            i.load(data)
            i.commit()
        '''
        for key in data:
            if data[key] is None:
                continue
            if key == 'ipaddr':
                for addr in self['ipaddr']:
                    self.del_ip(*addr)
                for addr in data[key]:
                    if isinstance(addr, basestring):
                        addr = (addr,)
                    self.add_ip(*addr)
            elif key == 'ports':
                for port in self['ports']:
                    self.del_port(port)
                for port in data[key]:
                    self.add_port(port)
            elif key == 'vlans':
                for vlan in self['vlans']:
                    self.del_vlan(vlan)
                for vlan in data[key]:
                    if vlan != 1:
                        self.add_vlan(vlan)
            elif key in ('neighbours', 'family'):
                # ignore on load
                pass
            else:
                self[key] = data[key]
        return self

    def load_dict(self, data):
        '''
        Update the interface info from a dictionary.

        This call always bypasses open transactions, loading
        changes directly into the interface data.
        '''
        with self._direct_state:
            self.load(data)

    def load_netlink(self, dev):
        '''
        Update the interface info from RTM_NEWLINK message.

        This call always bypasses open transactions, loading
        changes directly into the interface data.
        '''
        global supported_kinds

        with self._direct_state:
            if self['ipdb_scope'] == 'locked':
                # do not touch locked interfaces
                return

            if self['ipdb_scope'] in ('shadow', 'create'):
                # ignore non-broadcast messages
                if dev['header']['sequence_number'] != 0:
                    return
                # ignore ghost RTM_NEWLINK messages
                if (config.kernel[0] < 3) and (
                    not dev.get_attr('IFLA_AF_SPEC')
                ):
                    return

            for name, value in dev.items():
                self[name] = value
            for cell in dev['attrs']:
                #
                # Parse on demand
                #
                # At that moment, being not referenced, the
                # NLA is not decoded (yet). Calling
                # `__getitem__()` on nla_slot triggers the
                # NLA decoding, if the nla is referenced:
                #
                norm = ifinfmsg.nla2name(cell[0])
                if norm not in self.cleanup:
                    self[norm] = cell[1]
            # load interface kind
            linkinfo = dev.get_attr('IFLA_LINKINFO')
            if linkinfo is not None:
                kind = linkinfo.get_attr('IFLA_INFO_KIND')
                if kind is not None:
                    self['kind'] = kind
                    if kind == 'vlan':
                        data = linkinfo.get_attr('IFLA_INFO_DATA')
                        self['vlan_id'] = data.get_attr('IFLA_VLAN_ID')
                        self['vlan_protocol'] = data.get_attr(
                            'IFLA_VLAN_PROTOCOL'
                        )
                        self['vlan_flags'] = data.get_attr(
                            'IFLA_VLAN_FLAGS', {}
                        ).get('flags', 0)
                    if kind in supported_kinds:
                        data = linkinfo.get_attr('IFLA_INFO_DATA') or {}
                        for nla in data.get('attrs', []):
                            norm = ifinfmsg.nla2name(nla[0])
                            self[norm] = nla[1]
            # load vlans
            if dev['family'] == AF_BRIDGE:
                spec = dev.get_attr('IFLA_AF_SPEC')
                if spec is not None:
                    vlans = spec.get_attrs('IFLA_BRIDGE_VLAN_INFO')
                    vmap = {}
                    for vlan in vlans:
                        vmap[vlan['vid']] = vlan
                    vids = set(vmap.keys())
                    # remove vids we do not have anymore
                    for vid in self['vlans'] - vids:
                        self.del_vlan(vid)
                    for vid in vids - self['vlans']:
                        self.add_vlan(vmap[vid])
                protinfo = dev.get_attr('IFLA_PROTINFO')
                if protinfo is not None:
                    for attr, value in protinfo['attrs']:
                        attr = attr[5:].lower()
                        self[attr] = value
            # the rest is possible only when interface
            # is used in IPDB, not standalone
            if self.ipdb is not None:
                self['ipaddr'] = self.ipdb.ipaddr[self['index']]
                self['neighbours'] = self.ipdb.neighbours[self['index']]
            # finally, cleanup all not needed
            for item in self.cleanup:
                if item in self:
                    del self[item]

            # AF_BRIDGE messages for bridges contain
            # IFLA_MASTER == self.index, we should fix it
            if self.get('master', None) == self['index']:
                self['master'] = None

            self['ipdb_scope'] = 'system'

    def wait_ip(self, *argv, **kwarg):
        return self['ipaddr'].wait_ip(*argv, **kwarg)

    @with_transaction
    def add_ip(self, ip, mask=None, broadcast=None, anycast=None, scope=None):
        '''
        Add IP address to an interface

        Address formats:

            with ipdb.interfaces.eth0 as i:
                i.add_ip('192.168.0.1', 24)
                i.add_ip('192.168.0.2/24')
                i.add_ip('192.168.0.3/255.255.255.0')
                i.add_ip('192.168.0.4/24',
                         broadcast='192.168.0.255',
                         scope=254)
        '''
        family = 0
        # split mask
        if mask is None:
            ip, mask = ip.split('/')

        if ip.find(':') > -1:
            family = AF_INET6
            # normalize IPv6 format
            ip = inet_ntop(AF_INET6, inet_pton(AF_INET6, ip))
        else:
            family = AF_INET

        if isinstance(mask, basestring):
            try:
                mask = int(mask, 0)
            except:
                mask = dqn2int(mask, family)

        # if it is a transaction or an interface update, apply the change
        self['ipaddr'].unlink((ip, mask))
        request = {}
        if broadcast is not None:
            request['broadcast'] = broadcast
        if anycast is not None:
            request['anycast'] = anycast
        if scope is not None:
            request['scope'] = scope
        self['ipaddr'].add((ip, mask), raw=request)

    @with_transaction
    def del_ip(self, ip, mask=None):
        '''
        Delete IP address from an interface
        '''
        if mask is None:
            ip, mask = ip.split('/')
            if mask.find('.') > -1:
                mask = dqn2int(mask)
            else:
                mask = int(mask, 0)
        # normalize the address
        if ip.find(':') > -1:
            ip = inet_ntop(AF_INET6, inet_pton(AF_INET6, ip))
        if (ip, mask) in self['ipaddr']:
            self['ipaddr'].unlink((ip, mask))
            self['ipaddr'].remove((ip, mask))

    @with_transaction
    def add_vlan(self, vlan, flags=None):
        if isinstance(vlan, dict):
            vid = vlan['vid']
        else:
            vid = vlan
            vlan = {'vid': vlan, 'flags': 0}
        self['vlans'].unlink(vid)
        self['vlans'].add(vid, raw=(vlan, flags))

    @with_transaction
    def del_vlan(self, vlan):
        if vlan in self['vlans']:
            self['vlans'].unlink(vlan)
            self['vlans'].remove(vlan)

    @with_transaction
    def add_port(self, port):
        '''
        Add port to a bridge or bonding
        '''
        ifindex = self._resolve_port(port)
        if not ifindex:
            self._delay_add_port.add(port)
        else:
            self['ports'].unlink(ifindex)
            self['ports'].add(ifindex)

    @with_transaction
    def del_port(self, port):
        '''
        Remove port from a bridge or bonding
        '''
        ifindex = self._resolve_port(port)
        if not ifindex:
            self._delay_del_port.add(port)
        else:
            self['ports'].unlink(ifindex)
            self['ports'].remove(ifindex)

    def reload(self):
        '''
        Reload interface information
        '''
        countdown = 3
        while countdown:
            links = self.nl.get_links(self['index'])
            if links:
                self.load_netlink(links[0])
                break
            else:
                countdown -= 1
                time.sleep(1)
        return self

    def review(self):
        ret = super(Interface, self).review()
        last = self.current_tx
        if self['ipdb_scope'] == 'create':
            ret['+ipaddr'] = last['ipaddr']
            ret['+ports'] = last['ports']
            ret['+vlans'] = last['vlans']
            del ret['ports']
            del ret['ipaddr']
            del ret['vlans']
        if last._delay_add_port:
            ports = set(['*%s' % x for x in last._delay_add_port])
            if '+ports' in ret:
                ret['+ports'] |= ports
            else:
                ret['+ports'] = ports
        if last._delay_del_port:
            ports = set(['*%s' % x for x in last._delay_del_port])
            if '-ports' in ret:
                ret['-ports'] |= ports
            else:
                ret['-ports'] = ports
        return ret

    def _run(self, cmd, *argv, **kwarg):
        try:
            return cmd(*argv, **kwarg)
        except Exception as error:
            if self.partial:
                self.errors.append(error)
                return []
            raise error

    def _resolve_port(self, port):
        # for now just a stupid resolver, will be
        # improved later with search by mac, etc.
        if isinstance(port, Interface):
            return port['index']
        else:
            return self.ipdb.interfaces.get(port, {}).get('index', None)

    def commit(
        self,
        tid=None,
        transaction=None,
        commit_phase=1,
        commit_mask=0xFF,
        newif=False,
    ):
        '''
        Commit transaction. In the case of exception all
        changes applied during commit will be reverted.
        '''

        if not commit_phase & commit_mask:
            return self

        error = None
        added = None
        removed = None
        drop = self.ipdb.txdrop
        notx = True

        init = None
        debug = {'traceback': None, 'transaction': None, 'next_stage': None}

        if tid or transaction:
            notx = False

        if tid:
            transaction = self.global_tx[tid]
        else:
            transaction = transaction or self.current_tx

        if transaction.partial:
            transaction.errors = []

        with self._write_lock:
            # if the interface does not exist, create it first ;)
            if self['ipdb_scope'] != 'system':
                # a special case: transition "create" -> "remove"
                if (
                    transaction['ipdb_scope'] == 'remove'
                    and self['ipdb_scope'] == 'create'
                ):
                    self.invalidate()
                    return self

                newif = True
                self.set_target('ipdb_scope', 'system')
                try:
                    # 8<---------------------------------------------------
                    # link resolve
                    if self._deferred_link:
                        link_key, link_obj = self._deferred_link
                        transaction[link_key] = self._resolve_port(link_obj)
                        self._deferred_link = None

                    # 8<----------------------------------------------------
                    # ACHTUNG: hack for old platforms
                    if self['address'] == '00:00:00:00:00:00':
                        with self._direct_state:
                            self['address'] = None
                            self['broadcast'] = None
                    # 8<----------------------------------------------------
                    init = self.pick()
                    try:
                        request = {
                            key: transaction[key]
                            for key in filter(
                                lambda x: x[:5] != 'bond_'
                                and x[:7] != 'brport_'
                                and x[:3] != 'br_',
                                transaction,
                            )
                            if transaction[key] is not None
                        }
                        for key in ('net_ns_fd', 'net_ns_pid'):
                            if key in request:
                                with self._direct_state:
                                    self[key] = None
                                    del request[key]
                        self.nl.link('add', **request)
                    except NetlinkError as x:
                        # File exists
                        if x.code == errno.EEXIST:
                            # A bit special case, could be one of two cases:
                            #
                            # 1. A race condition between two different IPDB
                            #    processes
                            # 2. An attempt to create dummy0, gre0, bond0 when
                            #    the corrseponding module is not loaded. Being
                            #    loaded, the module creates a default interface
                            #    by itself, causing the request to fail
                            #
                            # The exception in that case can cause the DB
                            # inconsistence, since there can be queued not only
                            # the interface creation, but also IP address
                            # changes etc.
                            #
                            # So we ignore this particular exception and try to
                            # continue, as it is created by us.
                            #
                            # 3. An attempt to create VLAN or VXLAN interface
                            #    with the same ID but under different name
                            #
                            # In that case we should forward error properly
                            if self['kind'] in ('vlan', 'vxlan'):
                                newif = x

                        else:
                            raise
                except Exception as e:
                    if transaction.partial:
                        transaction.errors.append(e)
                        raise PartialCommitException()
                    else:
                        # If link('add', ...) raises an exception, no netlink
                        # broadcast will be sent, and the object is unmodified.
                        # After the exception forwarding, the object is ready
                        # to repeat the commit() call.
                        if drop and notx:
                            self.drop(transaction.uid)
                        raise

        if transaction['ipdb_scope'] == 'create' and commit_phase > 1:
            if self['index']:
                wd = self.ipdb.watchdog('RTM_DELLINK', ifname=self['ifname'])
                with self._direct_state:
                    self['ipdb_scope'] = 'locked'
                self.nl.link('delete', index=self['index'])
                wd.wait()
            self.load_dict(transaction)
            return self

        elif newif:
            # Here we come only if a new interface is created
            #
            if commit_phase == 1 and not self.wait_target('ipdb_scope'):
                if drop and notx:
                    self.drop(transaction.uid)
                self.invalidate()
                if isinstance(newif, Exception):
                    raise newif
                else:
                    raise CreateException()

            # Re-populate transaction.ipaddr to have a proper IP target
            #
            # The reason behind the code is that a new interface in the
            # "up" state will have automatic IPv6 addresses, that aren't
            # reflected in the transaction. This may cause a false IP
            # target mismatch and a commit failure.
            #
            # To avoid that, collect automatic addresses to the
            # transaction manually, since it is not yet properly linked.
            #
            for addr in self.ipdb.ipaddr[self['index']]:
                transaction['ipaddr'].add(addr)

            # Reload the interface data
            try:
                self.load_netlink(self.nl.link('get', **request)[0])
            except Exception:
                pass

        # now we have our index and IP set and all other stuff
        snapshot = self.pick()

        # make snapshots of all dependent routes
        if commit_phase == 1 and hasattr(self.ipdb, 'routes'):
            self.routes = []
            for record in self.ipdb.routes.filter({'oif': self['index']}):
                # For MPLS routes the key is an integer
                # They should match anyways
                if getattr(record['key'], 'table', None) != 255:
                    self.routes.append(
                        (record['route'], record['route'].pick())
                    )

        # resolve all delayed ports
        def resolve_ports(transaction, ports, callback, self, drop):
            def error(x):
                return KeyError('can not resolve port %s' % x)

            for port in tuple(ports):
                ifindex = self._resolve_port(port)
                if not ifindex:
                    if transaction.partial:
                        transaction.errors.append(error(port))
                    else:
                        if drop:
                            self.drop(transaction.uid)
                        raise error(port)
                else:
                    ports.remove(port)
                    with transaction._direct_state:  # ????
                        callback(ifindex)

        resolve_ports(
            transaction,
            transaction._delay_add_port,
            transaction.add_port,
            self,
            drop and notx,
        )
        resolve_ports(
            transaction,
            transaction._delay_del_port,
            transaction.del_port,
            self,
            drop and notx,
        )

        try:
            removed, added = snapshot // transaction

            run = transaction._run
            nl = transaction.nl

            # 8<---------------------------------------------
            # Port vlans
            if removed['vlans'] or added['vlans']:
                self['vlans'].set_target(transaction['vlans'])

                for i in removed['vlans']:
                    # remove vlan from the port
                    run(
                        nl.vlan_filter,
                        'del',
                        index=self['index'],
                        vlan_info=self['vlans'][i][0],
                    )

                for i in added['vlans']:
                    # add vlan to the port
                    vinfo = transaction['vlans'][i][0]
                    flags = transaction['vlans'][i][1]
                    req = {'index': self['index'], 'vlan_info': vinfo}
                    if flags == 'self':
                        req['vlan_flags'] = flags
                        # this request will NOT give echo,
                        # so bypass the check
                        with self._direct_state:
                            self.add_vlan(vinfo['vid'])
                    run(nl.vlan_filter, 'add', **req)

                self['vlans'].target.wait(SYNC_TIMEOUT)
                if not self['vlans'].target.is_set():
                    raise CommitException('vlans target is not set')

            # 8<---------------------------------------------
            # Ports
            if removed['ports'] or added['ports']:
                self['ports'].set_target(transaction['ports'])

                for i in removed['ports']:
                    # detach port
                    if i in self.ipdb.interfaces:
                        (
                            self.ipdb.interfaces[i]
                            .set_target('master', None)
                            .mirror_target('master', 'link')
                        )
                        run(nl.link, 'update', index=i, master=0)
                    else:
                        transaction.errors.append(KeyError(i))

                for i in added['ports']:
                    # attach port
                    if i in self.ipdb.interfaces:
                        (
                            self.ipdb.interfaces[i]
                            .set_target('master', self['index'])
                            .mirror_target('master', 'link')
                        )
                        run(nl.link, 'update', index=i, master=self['index'])
                    else:
                        transaction.errors.append(KeyError(i))

                self['ports'].target.wait(SYNC_TIMEOUT)
                if self['ports'].target.is_set():
                    for msg in self.nl.get_vlans(index=self['index']):
                        self.load_netlink(msg)
                else:
                    raise CommitException('ports target is not set')

                # 1. wait for proper targets on ports
                # 2. wait for mtu sync
                #
                # the bridge mtu is set from the port, if the latter is smaller
                # the bond mtu sets the port mtu, if the latter is smaller
                #
                # FIXME: team interfaces?
                for i in list(added['ports']) + list(removed['ports']):
                    port = self.ipdb.interfaces[i]
                    # port update
                    target = port._local_targets['master']
                    target.wait(SYNC_TIMEOUT)
                    with port._write_lock:
                        del port._local_targets['master']
                        del port._local_targets['link']
                    if not target.is_set():
                        raise CommitException('master target failed')
                    if i in added['ports']:
                        if port.if_master != self['index']:
                            raise CommitException('master set failed')
                    else:
                        if port.if_master == self['index']:
                            raise CommitException('master unset failed')
                    # master update
                    if self['kind'] == 'bridge' and self['mtu'] > port['mtu']:
                        self.set_target('mtu', port['mtu'])
                        self.wait_target('mtu')

            # 8<---------------------------------------------
            # Interface changes
            request = {}
            brequest = {}
            prequest = {}
            # preseed requests with the interface kind
            request['kind'] = self['kind']
            brequest['kind'] = self['kind']
            wait_all = False
            for key, value in added.items():
                if (
                    value is not None
                    and (key not in self._virtual_fields)
                    and (key != 'kind')
                ):
                    if key[:3] == 'br_':
                        brequest[key] = added[key]
                    elif key[:7] == 'brport_':
                        prequest[key[7:]] = added[key]
                    else:
                        if key == 'address' and added[key] is not None:
                            self[key] = added[key].lower()
                        request[key] = added[key]
            # FIXME: flush the interface type so the next two conditions
            # will work correctly
            request['kind'] = None
            brequest['kind'] = None

            # apply changes only if there is something to apply
            if (self['kind'] == 'bridge') and any(
                [brequest[item] is not None for item in brequest]
            ):
                brequest['index'] = self['index']
                brequest['kind'] = self['kind']
                brequest['family'] = AF_BRIDGE
                wait_all = True
                run(nl.link, 'set', **brequest)

            if any([request[item] is not None for item in request]):
                request['index'] = self['index']
                request['kind'] = self['kind']
                if request.get('address', None) == '00:00:00:00:00:00':
                    request.pop('address')
                    request.pop('broadcast', None)
                wait_all = True
                run(nl.link, 'update', **request)
                # Yet another trick: setting ifalias doesn't cause
                # netlink updates
                if 'ifalias' in request:
                    self.reload()

            if any([prequest[item] is not None for item in prequest]):
                prequest['index'] = self['index']
                run(nl.brport, 'set', **prequest)

            if (wait_all) and (not transaction.partial):
                transaction.wait_all_targets()

            # 8<---------------------------------------------
            # VLAN flags -- a dirty hack, pls do something with it
            if added.get('vlan_flags') is not None:
                run(
                    nl.link,
                    'set',
                    **{
                        'kind': 'vlan',
                        'index': self['index'],
                        'vlan_flags': added['vlan_flags'],
                    }
                )

            # 8<---------------------------------------------
            # IP address changes
            for _ in range(3):
                ip2add = transaction['ipaddr'] - self['ipaddr']
                ip2remove = self['ipaddr'] - transaction['ipaddr']

                if not ip2add and not ip2remove:
                    break

                self['ipaddr'].set_target(transaction['ipaddr'])
                ###
                # Remove
                #
                # The promote_secondaries sysctl causes the kernel
                # to add secondary addresses back after the primary
                # address is removed.
                #
                # The library can not tell this from the result of
                # an external program.
                #
                # One simple way to work that around is to remove
                # secondaries first.
                rip = sorted(
                    ip2remove,
                    key=lambda x: self['ipaddr'][x]['flags'],
                    reverse=True,
                )
                # 8<--------------------------------------
                for i in rip:
                    # When you remove a primary IP addr, all the
                    # subnetwork can be removed. In this case you
                    # will fail, but it is OK, no need to roll back
                    try:
                        run(
                            nl.addr,
                            'delete',
                            index=self['index'],
                            address=i[0],
                            prefixlen=i[1],
                        )
                    except NetlinkError as x:
                        # bypass only errno 99,
                        # 'Cannot assign address'
                        if x.code != errno.EADDRNOTAVAIL:
                            raise
                    except socket_error as x:
                        # bypass illegal IP requests
                        if isinstance(x.args[0], basestring) and x.args[
                            0
                        ].startswith('illegal IP'):
                            continue
                        raise
                ###
                # Add addresses
                # 8<--------------------------------------
                for i in ip2add:
                    # Try to fetch additional address attributes
                    try:
                        kwarg = dict(
                            [
                                k
                                for k in transaction['ipaddr'][i].items()
                                if k[0] in ('broadcast', 'anycast', 'scope')
                            ]
                        )
                    except KeyError:
                        kwarg = None
                    try:
                        # feed the address to the OS
                        kwarg = kwarg or {}
                        kwarg['index'] = self['index']
                        kwarg['address'] = i[0]
                        kwarg['prefixlen'] = i[1]
                        run(nl.addr, 'add', **kwarg)
                    except NetlinkError as x:
                        if x.code != errno.EEXIST:
                            raise

                # 8<--------------------------------------
                # some interfaces do not send IPv6 address
                # updates, when are down
                #
                # beside of that, bridge interfaces are
                # down by default, so they never send
                # address updates from beginning
                #
                # FIXME:
                #
                # that all is a dirtiest hack ever, pls do
                # something with it
                #
                if (not self['flags'] & 1) or hasattr(self.ipdb.nl, 'netns'):
                    # 1. flush old IPv6 addresses
                    for addr in list(self['ipaddr'].ipv6):
                        self['ipaddr'].remove(addr)
                    # 2. reload addresses
                    for addr in self.nl.get_addr(
                        index=self['index'], family=AF_INET6
                    ):
                        self.ipdb.ipaddr._new(addr)
                    # if there are tons of IPv6 addresses, it may take a
                    # really long time, and that's bad, but it's broken in
                    # the kernel :|

                # 8<--------------------------------------
                self['ipaddr'].target.wait(SYNC_TIMEOUT)
                if self['ipaddr'].target.is_set():
                    break
            else:
                raise CommitException('ipaddr target is not set')

            # 8<---------------------------------------------
            # Iterate callback chain
            for ch in self._commit_hooks:
                # An exception will rollback the transaction
                ch(self.dump(), snapshot.dump(), transaction.dump())

            # 8<---------------------------------------------
            # Move the interface to a netns
            if ('net_ns_fd' in added) or ('net_ns_pid' in added):
                request = {}
                for key in ('net_ns_fd', 'net_ns_pid'):
                    if key in added:
                        request[key] = added[key]

                request['index'] = self['index']
                run(nl.link, 'update', **request)

                countdown = 10
                while countdown:
                    # wait until the interface will disappear
                    # from the current network namespace --
                    # up to 1 second (make it configurable?)
                    try:
                        self.nl.get_links(self['index'])
                    except NetlinkError as e:
                        if e.code == errno.ENODEV:
                            break
                        raise
                    except Exception:
                        raise
                    countdown -= 1
                    time.sleep(0.1)

            # 8<---------------------------------------------
            # Interface removal
            if added.get('ipdb_scope') in ('shadow', 'remove'):
                wd = self.ipdb.watchdog('RTM_DELLINK', ifname=self['ifname'])
                with self._direct_state:
                    self['ipdb_scope'] = 'locked'
                self.nl.link('delete', index=self['index'])
                wd.wait()

                with self._direct_state:
                    self['ipdb_scope'] = 'shadow'

                # system-wide checks
                if commit_phase == 1:
                    self.ipdb.ensure('run')

                if added.get('ipdb_scope') == 'remove':
                    self.ipdb.interfaces._detach(None, self['index'], None)

                if notx:
                    self.drop(transaction.uid)

                return self
            # 8<---------------------------------------------

            # system-wide checks
            if commit_phase == 1:
                self.ipdb.ensure('run')

            # so far all's ok
            drop = True

        except Exception as e:
            error = e
            # log the error environment
            debug['traceback'] = traceback.format_exc()
            debug['transaction'] = transaction
            debug['next_stage'] = None

            # something went wrong: roll the transaction back
            if commit_phase == 1:
                if newif:
                    drop = False
                try:
                    self.commit(
                        transaction=init if newif else snapshot,
                        commit_phase=2,
                        commit_mask=commit_mask,
                        newif=newif,
                    )

                except Exception as i_e:
                    debug['next_stage'] = i_e
                    error = RuntimeError()
            else:
                # reload all the database -- it can take a long time,
                # but it is required since we have no idea, what is
                # the result of the failure
                links = self.nl.get_links()
                for link in links:
                    self.ipdb.interfaces._new(link)
                links = self.nl.get_vlans()
                for link in links:
                    self.ipdb.interfaces._new(link)
                for addr in self.nl.get_addr():
                    self.ipdb.ipaddr._new(addr)

            for key in ('ipaddr', 'ports', 'vlans'):
                self[key].clear_target()

        # raise partial commit exceptions
        if transaction.partial and transaction.errors:
            error = PartialCommitException('partial commit error')

        # drop only if required
        if drop and notx:
            # drop last transaction in any case
            self.drop(transaction.uid)

        # raise exception for failed transaction
        if error is not None:
            error.debug = debug
            raise error

        # restore dependent routes for successful rollback
        if commit_phase == 2:
            for route in self.routes:
                with route[0]._direct_state:
                    route[0]['ipdb_scope'] = 'restore'
                try:
                    route[0].commit(
                        transaction=route[1], commit_phase=2, commit_mask=2
                    )
                except RuntimeError as x:
                    # RuntimeError is raised due to phase 2, so
                    # an additional check is required
                    if (
                        isinstance(x.cause, NetlinkError)
                        and x.cause.code == errno.EEXIST
                    ):
                        pass

        time.sleep(config.commit_barrier)

        # drop all collected errors, if any
        self.errors = []
        return self

    def up(self):
        '''
        Shortcut: change the interface state to 'up'.
        '''
        self['state'] = 'up'
        return self

    def down(self):
        '''
        Shortcut: change the interface state to 'down'.
        '''
        self['state'] = 'down'
        return self

    def remove(self):
        '''
        Mark the interface for removal
        '''
        self['ipdb_scope'] = 'remove'
        return self

    def shadow(self):
        '''
        Remove the interface from the OS, but leave it in the
        database. When one will try to re-create interface with
        the same name, all the old saved attributes will apply
        to the new interface, incl. MAC-address and even the
        interface index. Please be aware, that the interface
        index can be reused by OS while the interface is "in the
        shadow state", in this case re-creation will fail.
        '''
        self['ipdb_scope'] = 'shadow'
        return self


class InterfacesDict(Dotkeys):
    def __init__(self, ipdb):
        self.ipdb = ipdb
        self._event_map = {'RTM_NEWLINK': self._new, 'RTM_DELLINK': self._del}

    def _register(self):
        links = self.ipdb.nl.get_links()
        # iterate twice to map port/master relations
        for link in links:
            self._new(link, skip_master=True)
        for link in links:
            self._new(link)
        # load bridge vlan information
        links = self.ipdb.nl.get_vlans()
        for link in links:
            self._new(link)

    def add(self, kind, ifname, reuse=False, **kwarg):
        '''
        Create new network interface
        '''
        with self.ipdb.exclusive:
            # check for existing interface
            if ifname in self:
                if (self[ifname]['ipdb_scope'] == 'shadow') or reuse:
                    device = self[ifname]
                    kwarg['kind'] = kind
                    device.load_dict(kwarg)
                    if self[ifname]['ipdb_scope'] == 'shadow':
                        with device._direct_state:
                            device['ipdb_scope'] = 'create'
                    device.begin()
                else:
                    raise CreateException("interface %s exists" % ifname)
            else:
                device = self[ifname] = Interface(
                    ipdb=self.ipdb, mode='snapshot'
                )
                # delay link resolve?
                for key in kwarg:
                    # any /.+link$/ attr
                    if key[-4:] == 'link':
                        if isinstance(kwarg[key], Interface):
                            kwarg[key] = kwarg[key].get('index') or kwarg[
                                key
                            ].get('ifname')
                        if not isinstance(kwarg[key], int):
                            device._deferred_link = (key, kwarg[key])
                device._mode = self.ipdb.mode
                with device._direct_state:
                    device['kind'] = kind
                    device['index'] = kwarg.get('index', 0)
                    device['ifname'] = ifname
                    device['ipdb_scope'] = 'create'
                    # set some specific attrs
                    for attr in (
                        'peer',
                        'uid',
                        'gid',
                        'ifr',
                        'mode',
                        'bond_mode',
                        'address',
                    ):
                        if attr in kwarg:
                            device[attr] = kwarg.pop(attr)
                device.begin()
                device.load(kwarg)
        return device

    def _del(self, msg):
        target = self.get(msg['index'])
        if target is None:
            return

        if msg['family'] == AF_BRIDGE:
            with target._direct_state:
                for vlan in tuple(target['vlans']):
                    target.del_vlan(vlan)

        # check for freezed devices
        if getattr(target, '_freeze', None):
            with target._direct_state:
                target['ipdb_scope'] = 'shadow'
            return

        # check for locked devices
        if target.get('ipdb_scope') in ('locked', 'shadow'):
            return

        self._detach(None, msg['index'], msg)

    def _new(self, msg, skip_master=False):
        # check, if a record exists
        index = msg.get('index', None)
        ifname = msg.get_attr('IFLA_IFNAME', None)
        device = None
        cleanup = None

        # scenario #1: no matches for both: new interface
        #
        # scenario #2: ifname exists, index doesn't:
        #              index changed
        # scenario #3: index exists, ifname doesn't:
        #              name changed
        # scenario #4: both exist: assume simple update and
        #              an optional name change

        if (index not in self) and (ifname not in self):
            # scenario #1, new interface
            device = self[index] = self[ifname] = Interface(ipdb=self.ipdb)
        elif (index not in self) and (ifname in self):
            # scenario #2, index change
            old_index = self[ifname]['index']
            device = self[index] = self[ifname]
            if old_index in self:
                cleanup = old_index

            if old_index in self.ipdb.ipaddr:
                self.ipdb.ipaddr[index] = self.ipdb.ipaddr[old_index]
                del self.ipdb.ipaddr[old_index]

            if old_index in self.ipdb.neighbours:
                self.ipdb.neighbours[index] = self.ipdb.neighbours[old_index]
                del self.ipdb.neighbours[old_index]
        else:
            # scenario #3, interface rename
            # scenario #4, assume rename
            old_name = self[index]['ifname']
            if old_name != ifname:
                # unlink old name
                cleanup = old_name
            device = self[ifname] = self[index]

        if index not in self.ipdb.ipaddr:
            self.ipdb.ipaddr[index] = self.ipdb._ipaddr_set()

        if index not in self.ipdb.neighbours:
            self.ipdb.neighbours[index] = LinkedSet()

        # update port references
        old_master = device.get('master', None)
        new_master = msg.get_attr('IFLA_MASTER')

        if old_master != new_master:
            if old_master in self:
                with self[old_master]._direct_state:
                    if index in self[old_master]['ports']:
                        self[old_master].del_port(index)
            if new_master in self and new_master != index:
                with self[new_master]._direct_state:
                    self[new_master].add_port(index)

        if cleanup is not None:
            del self[cleanup]

        if skip_master:
            msg.strip('IFLA_MASTER')

        device.load_netlink(msg)
        if new_master is None:
            with device._direct_state:
                device['master'] = None

    def _detach(self, name, idx, msg=None):
        with self.ipdb.exclusive:
            if msg is not None:
                if (
                    msg['event'] == 'RTM_DELLINK'
                    and msg['change'] != 0xFFFFFFFF
                ):
                    return
            if idx is None or idx < 1:
                target = self[name]
                idx = target['index']
            else:
                target = self[idx]
                name = target['ifname']
            # clean up port, if exists
            master = target.get('master', None)
            if master in self and target['index'] in self[master]['ports']:
                with self[master]._direct_state:
                    self[master].del_port(target)
            self.pop(name, None)
            self.pop(idx, None)
            self.ipdb.ipaddr.pop(idx, None)
            self.ipdb.neighbours.pop(idx, None)
            with target._direct_state:
                target['ipdb_scope'] = 'detached'


class AddressesDict(dict):
    def __init__(self, ipdb):
        self.ipdb = ipdb
        self._event_map = {'RTM_NEWADDR': self._new, 'RTM_DELADDR': self._del}

    def _register(self):
        for msg in self.ipdb.nl.get_addr():
            self._new(msg)

    def reload(self):
        # Reload addresses from the kernel.
        # (This is a workaround to reorder primary and secondary addresses.)
        for k in self.keys():
            self[k] = self.ipdb._ipaddr_set()
        for msg in self.ipdb.nl.get_addr():
            self._new(msg)
        for idx in self.keys():
            iff = self.ipdb.interfaces[idx]
            with iff._direct_state:
                iff['ipaddr'] = self[idx]

    def _new(self, msg):
        if msg['family'] == AF_INET:
            addr = msg.get_attr('IFA_LOCAL')
        elif msg['family'] == AF_INET6:
            addr = msg.get_attr('IFA_LOCAL')
            if not addr:
                addr = msg.get_attr('IFA_ADDRESS')
        else:
            return
        raw = {
            'local': msg.get_attr('IFA_LOCAL'),
            'broadcast': msg.get_attr('IFA_BROADCAST'),
            'address': msg.get_attr('IFA_ADDRESS'),
            'flags': msg.get_attr('IFA_FLAGS') or msg.get('flags'),
            'prefixlen': msg['prefixlen'],
            'family': msg['family'],
            'cacheinfo': msg.get_attr('IFA_CACHEINFO'),
        }
        try:
            self[msg['index']].add(key=(addr, raw['prefixlen']), raw=raw)
        except:
            pass

    def _del(self, msg):
        if msg['family'] == AF_INET:
            addr = msg.get_attr('IFA_LOCAL')
        elif msg['family'] == AF_INET6:
            addr = msg.get_attr('IFA_ADDRESS')
        else:
            return
        try:
            self[msg['index']].remove((addr, msg['prefixlen']))
        except:
            pass


class NeighboursDict(dict):
    def __init__(self, ipdb):
        self.ipdb = ipdb
        self._event_map = {
            'RTM_NEWNEIGH': self._new,
            'RTM_DELNEIGH': self._del,
        }

    def _register(self):
        for msg in self.ipdb.nl.get_neighbours():
            self._new(msg)

    def _new(self, msg):
        if msg['family'] == AF_BRIDGE:
            return

        try:
            (
                self[msg['ifindex']].add(
                    key=msg.get_attr('NDA_DST'),
                    raw={'lladdr': msg.get_attr('NDA_LLADDR')},
                )
            )
        except:
            pass

    def _del(self, msg):
        if msg['family'] == AF_BRIDGE:
            return
        try:
            (self[msg['ifindex']].remove(msg.get_attr('NDA_DST')))
        except:
            pass


spec = [
    {'name': 'interfaces', 'class': InterfacesDict, 'kwarg': {}},
    {
        'name': 'by_name',
        'class': View,
        'kwarg': {
            'path': 'interfaces',
            'constraint': lambda k, v: isinstance(k, basestring),
        },
    },
    {
        'name': 'by_index',
        'class': View,
        'kwarg': {
            'path': 'interfaces',
            'constraint': lambda k, v: isinstance(k, int),
        },
    },
    {'name': 'ipaddr', 'class': AddressesDict, 'kwarg': {}},
    {'name': 'neighbours', 'class': NeighboursDict, 'kwarg': {}},
]



from socket import AF_INET, AF_INET6, AF_UNIX, IPPROTO_TCP, inet_ntop
from struct import pack

from pyroute2.netlink import (
    NETLINK_SOCK_DIAG,
    NLM_F_MATCH,
    NLM_F_REQUEST,
    NLM_F_ROOT,
    nla,
    nlmsg,
)
from pyroute2.netlink.nlsocket import Marshal, NetlinkSocket

SOCK_DIAG_BY_FAMILY = 20
SOCK_DESTROY = 21

# states
SS_UNKNOWN = 0
SS_ESTABLISHED = 1
SS_SYN_SENT = 2
SS_SYN_RECV = 3
SS_FIN_WAIT1 = 4
SS_FIN_WAIT2 = 5
SS_TIME_WAIT = 6
SS_CLOSE = 7
SS_CLOSE_WAIT = 8
SS_LAST_ACK = 9
SS_LISTEN = 10
SS_CLOSING = 11
SS_MAX = 12

SS_ALL = (1 << SS_MAX) - 1
SS_CONN = SS_ALL & ~(
    (1 << SS_LISTEN)
    | (1 << SS_CLOSE)
    | (1 << SS_TIME_WAIT)
    | (1 << SS_SYN_RECV)
)

# multicast groups ids (for use with {add,drop}_membership)
SKNLGRP_NONE = 0
SKNLGRP_INET_TCP_DESTROY = 1
SKNLGRP_INET_UDP_DESTROY = 2
SKNLGRP_INET6_TCP_DESTROY = 3
SKNLGRP_INET6_UDP_DESTROY = 4


class sock_diag_req(nlmsg):
    fields = (('sdiag_family', 'B'), ('sdiag_protocol', 'B'))


UDIAG_SHOW_NAME = 0x01
UDIAG_SHOW_VFS = 0x02
UDIAG_SHOW_PEER = 0x04
UDIAG_SHOW_ICONS = 0x08
UDIAG_SHOW_RQLEN = 0x10
UDIAG_SHOW_MEMINFO = 0x20


class inet_addr_codec(nlmsg):
    def encode(self):
        # FIXME: add human-friendly API to specify IP addresses as str
        # (see also decode())
        if self['idiag_src'] == 0:
            self['idiag_src'] = (0, 0, 0, 0)
        if self['idiag_dst'] == 0:
            self['idiag_dst'] = (0, 0, 0, 0)
        nlmsg.encode(self)

    def decode(self):
        nlmsg.decode(self)
        if self[self.ffname] == AF_INET:
            self['idiag_dst'] = inet_ntop(
                AF_INET, pack('>I', self['idiag_dst'][0])
            )
            self['idiag_src'] = inet_ntop(
                AF_INET, pack('>I', self['idiag_src'][0])
            )
        elif self[self.ffname] == AF_INET6:
            self['idiag_dst'] = inet_ntop(
                AF_INET6, pack('>IIII', *self['idiag_dst'])
            )
            self['idiag_src'] = inet_ntop(
                AF_INET6, pack('>IIII', *self['idiag_src'])
            )


class inet_diag_req(inet_addr_codec):
    ffname = 'sdiag_family'
    fields = (
        ('sdiag_family', 'B'),
        ('sdiag_protocol', 'B'),
        ('idiag_ext', 'B'),
        ('__pad', 'B'),
        ('idiag_states', 'I'),
        ('idiag_sport', '>H'),
        ('idiag_dport', '>H'),
        ('idiag_src', '>4I'),
        ('idiag_dst', '>4I'),
        ('idiag_if', 'I'),
        ('idiag_cookie', 'Q'),
    )


class inet_diag_msg(inet_addr_codec):
    ffname = 'idiag_family'
    fields = (
        ('idiag_family', 'B'),
        ('idiag_state', 'B'),
        ('idiag_timer', 'B'),
        ('idiag_retrans', 'B'),
        ('idiag_sport', '>H'),
        ('idiag_dport', '>H'),
        ('idiag_src', '>4I'),
        ('idiag_dst', '>4I'),
        ('idiag_if', 'I'),
        ('idiag_cookie', 'Q'),
        ('idiag_expires', 'I'),
        ('idiag_rqueue', 'I'),
        ('idiag_wqueue', 'I'),
        ('idiag_uid', 'I'),
        ('idiag_inode', 'I'),
    )

    nla_map = (
        ('INET_DIAG_NONE', 'none'),
        ('INET_DIAG_MEMINFO', 'inet_diag_meminfo'),
        # FIXME: must be protocol specific?
        ('INET_DIAG_INFO', 'tcp_info'),
        ('INET_DIAG_VEGASINFO', 'tcpvegas_info'),
        ('INET_DIAG_CONG', 'asciiz'),
        ('INET_DIAG_TOS', 'hex'),
        ('INET_DIAG_TCLASS', 'hex'),
        ('INET_DIAG_SKMEMINFO', 'hex'),
        ('INET_DIAG_SHUTDOWN', 'uint8'),
        ('INET_DIAG_DCTCPINFO', 'tcp_dctcp_info'),
        ('INET_DIAG_PROTOCOL', 'hex'),
        ('INET_DIAG_SKV6ONLY', 'uint8'),
        ('INET_DIAG_LOCALS', 'hex'),
        ('INET_DIAG_PEERS', 'hex'),
        ('INET_DIAG_PAD', 'hex'),
        ('INET_DIAG_MARK', 'hex'),
        ('INET_DIAG_BBRINFO', 'tcp_bbr_info'),
        ('INET_DIAG_CLASS_ID', 'uint32'),
        ('INET_DIAG_MD5SIG', 'hex'),
        ('INET_DIAG_ULP_INFO', 'hex'),
        ('INET_DIAG_SK_BPF_STORAGES', 'hex'),
        ('INET_DIAG_CGROUP_ID', 'uint64'),
    )

    class inet_diag_meminfo(nla):
        fields = (
            ('idiag_rmem', 'I'),
            ('idiag_wmem', 'I'),
            ('idiag_fmem', 'I'),
            ('idiag_tmem', 'I'),
        )

    class tcpvegas_info(nla):
        fields = (
            ('tcpv_enabled', 'I'),
            ('tcpv_rttcnt', 'I'),
            ('tcpv_rtt', 'I'),
            ('tcpv_minrtt', 'I'),
        )

    class tcp_dctcp_info(nla):
        fields = (
            ('dctcp_enabled', 'H'),
            ('dctcp_ce_state', 'H'),
            ('dctcp_alpha', 'I'),
            ('dctcp_ab_ecn', 'I'),
            ('dctcp_ab_tot', 'I'),
        )

    class tcp_bbr_info(nla):
        fields = (
            ('bbr_bw_lo', 'I'),
            ('bbr_bw_hi', 'I'),
            ('bbr_min_rtt', 'I'),
            ('bbr_pacing_gain', 'I'),
            ('bbr_cwnd_gain', 'I'),
        )

    class tcp_info(nla):
        fields = (
            ('tcpi_state', 'B'),
            ('tcpi_ca_state', 'B'),
            ('tcpi_retransmits', 'B'),
            ('tcpi_probes', 'B'),
            ('tcpi_backoff', 'B'),
            ('tcpi_options', 'B'),
            ('tcpi_snd_wscale', 'B'),  # tcpi_rcv_wscale -- in decode()
            ('tcpi_delivery_rate_app_limited', 'B'),
            ('tcpi_rto', 'I'),
            ('tcpi_ato', 'I'),
            ('tcpi_snd_mss', 'I'),
            ('tcpi_rcv_mss', 'I'),
            ('tcpi_unacked', 'I'),
            ('tcpi_sacked', 'I'),
            ('tcpi_lost', 'I'),
            ('tcpi_retrans', 'I'),
            ('tcpi_fackets', 'I'),
            # Times
            ('tcpi_last_data_sent', 'I'),
            ('tcpi_last_ack_sent', 'I'),
            ('tcpi_last_data_recv', 'I'),
            ('tcpi_last_ack_recv', 'I'),
            # Metrics
            ('tcpi_pmtu', 'I'),
            ('tcpi_rcv_ssthresh', 'I'),
            ('tcpi_rtt', 'I'),
            ('tcpi_rttvar', 'I'),
            ('tcpi_snd_ssthresh', 'I'),
            ('tcpi_snd_cwnd', 'I'),
            ('tcpi_advmss', 'I'),
            ('tcpi_reordering', 'I'),
            ('tcpi_rcv_rtt', 'I'),
            ('tcpi_rcv_space', 'I'),
            ('tcpi_total_retrans', 'I'),
            ('tcpi_pacing_rate', 'Q'),
            ('tcpi_max_pacing_rate', 'Q'),
            ('tcpi_bytes_acked', 'Q'),
            ('tcpi_bytes_received', 'Q'),
            ('tcpi_segs_out', 'I'),
            ('tcpi_segs_in', 'I'),
            ('tcpi_notsent_bytes', 'I'),
            ('tcpi_min_rtt', 'I'),
            ('tcpi_data_segs_in', 'I'),
            ('tcpi_data_segs_out', 'I'),
            ('tcpi_delivery_rate', 'Q'),
            ('tcpi_busy_time', 'Q'),
            ('tcpi_rwnd_limited', 'Q'),
            ('tcpi_sndbuf_limited', 'Q'),
            ('tcpi_delivered', 'I'),
            ('tcpi_delivered_ce', 'I'),
            ('tcpi_bytes_sent', 'Q'),
            ('tcpi_bytes_retrans', 'Q'),
            ('tcpi_dsack_dups', 'I'),
            ('tcpi_reord_seen', 'I'),
            ('tcpi_rcv_ooopack', 'I'),
            ('tcpi_snd_wnd', 'I'),
        )

        def decode(self):
            # Fix tcpi_rcv_scale amd delivery_rate bit fields.
            # In the C:
            #
            # __u8    tcpi_snd_wscale : 4, tcpi_rcv_wscale : 4;
            # __u8    tcpi_delivery_rate_app_limited:1;
            #
            nla.decode(self)
            self['tcpi_rcv_wscale'] = self['tcpi_snd_wscale'] & 0xF
            self['tcpi_snd_wscale'] = self['tcpi_snd_wscale'] & 0xF0 >> 4
            self['tcpi_delivery_rate_app_limited'] = (
                self['tcpi_delivery_rate_app_limited'] & 0x80 >> 7
            )


class unix_diag_req(nlmsg):
    fields = (
        ('sdiag_family', 'B'),
        ('sdiag_protocol', 'B'),
        ('__pad', 'H'),
        ('udiag_states', 'I'),
        ('udiag_ino', 'I'),
        ('udiag_show', 'I'),
        ('udiag_cookie', 'Q'),
    )


class unix_diag_msg(nlmsg):
    fields = (
        ('udiag_family', 'B'),
        ('udiag_type', 'B'),
        ('udiag_state', 'B'),
        ('__pad', 'B'),
        ('udiag_ino', 'I'),
        ('udiag_cookie', 'Q'),
    )

    nla_map = (
        ('UNIX_DIAG_NAME', 'asciiz'),
        ('UNIX_DIAG_VFS', 'unix_diag_vfs'),
        ('UNIX_DIAG_PEER', 'uint32'),
        ('UNIX_DIAG_ICONS', 'hex'),
        ('UNIX_DIAG_RQLEN', 'unix_diag_rqlen'),
        ('UNIX_DIAG_MEMINFO', 'hex'),
        ('UNIX_DIAG_SHUTDOWN', 'uint8'),
    )

    class unix_diag_vfs(nla):
        fields = (('udiag_vfs_ino', 'I'), ('udiag_vfs_dev', 'I'))

    class unix_diag_rqlen(nla):
        fields = (('udiag_rqueue', 'I'), ('udiag_wqueue', 'I'))


class MarshalDiag(Marshal):
    key_format = 'B'
    # The family goes after the nlmsg header,
    # IHHII = 4 + 2 + 2 + 4 + 4 = 16 bytes
    key_offset = 16
    # Please notice that the SOCK_DIAG Marshal
    # uses not the nlmsg type, but sdiag_family
    # to choose the proper class
    msg_map = {
        AF_UNIX: unix_diag_msg,
        AF_INET: inet_diag_msg,
        AF_INET6: inet_diag_msg,
    }
    # error type NLMSG_ERROR == 2 == AF_INET,
    # it doesn't work for DiagSocket that way,
    # so disable the error messages for now
    error_type = -1


class DiagSocket(NetlinkSocket):
    '''
    Usage::

        from pyroute2 import DiagSocket
        with DiagSocket() as ds:
            ds.bind()
            sstats = ds.get_sock_stats()

    '''

    def __init__(self, fileno=None):
        super(DiagSocket, self).__init__(NETLINK_SOCK_DIAG)
        self.marshal = MarshalDiag()

    def get_sock_stats(
        self,
        family=AF_UNIX,
        states=SS_ALL,
        protocol=IPPROTO_TCP,
        extensions=0,
        show=(
            UDIAG_SHOW_NAME
            | UDIAG_SHOW_VFS
            | UDIAG_SHOW_PEER
            | UDIAG_SHOW_ICONS
        ),
    ):
        '''
        Get sockets statistics.

        ACHTUNG: the get_sock_stats() signature will be changed
        before the next release, this one is a WIP-code!
        '''

        if family == AF_UNIX:
            req = unix_diag_req()
            req['udiag_states'] = states
            req['udiag_show'] = show
        elif family in (AF_INET, AF_INET6):
            req = inet_diag_req()
            req['idiag_states'] = states
            req['sdiag_protocol'] = protocol
            req['idiag_ext'] = extensions
        else:
            raise NotImplementedError()
        req['sdiag_family'] = family

        return tuple(
            self.nlm_request(
                req,
                SOCK_DIAG_BY_FAMILY,
                NLM_F_REQUEST | NLM_F_ROOT | NLM_F_MATCH,
            )
        )



"""
Docker executor module

.. versionadded:: 2019.2.0

Used with the docker proxy minion.
"""

__virtualname__ = "docker"

DOCKER_MOD_MAP = {
    "state.sls": "docker.sls",
    "state.apply": "docker.apply",
    "state.highstate": "docker.highstate",
}

__deprecated__ = (
    3009,
    "docker",
    "https://github.com/saltstack/saltext-docker",
)


def __virtual__():
    if "proxy" not in __opts__:
        return (
            False,
            "Docker executor is only meant to be used with Docker Proxy Minions",
        )
    if __opts__.get("proxy", {}).get("proxytype") != __virtualname__:
        return False, f"Proxytype does not match: {__virtualname__}"
    return True


def execute(opts, data, func, args, kwargs):
    """
    Directly calls the given function with arguments
    """
    if data["fun"] == "saltutil.find_job":
        return __executors__["direct_call.execute"](opts, data, func, args, kwargs)
    if data["fun"] in DOCKER_MOD_MAP:
        return __executors__["direct_call.execute"](
            opts,
            data,
            __salt__[DOCKER_MOD_MAP[data["fun"]]],
            [opts["proxy"]["name"]] + args,
            kwargs,
        )
    return __salt__["docker.call"](opts["proxy"]["name"], data["fun"], *args, **kwargs)


def allow_missing_func(function):  # pylint: disable=unused-argument
    """
    Allow all calls to be passed through to docker container.

    The docker call will use direct_call, which will return back if the module
    was unable to be run.
    """
    return True



#-----------------------------------------------------------------------------
#   Copyright (c) 2008 by David P. D. Moss. All rights reserved.
#
#   Released under the BSD license. See the LICENSE file for details.
#-----------------------------------------------------------------------------
"""A basic implementation of RFC 1924 ;-)"""

from netaddr.core import AddrFormatError
from netaddr.ip import IPAddress

from netaddr.compat import _zip


def chr_range(low, high):
    """Returns all characters between low and high chars."""
    return [chr(i) for i in range(ord(low), ord(high) + 1)]

#: Base 85 integer index to character lookup table.
BASE_85 = (
    chr_range('0', '9') + chr_range('A', 'Z') +
    chr_range('a', 'z') +
    ['!', '#', '$', '%', '&', '(', ')', '*', '+', '-', ';', '<', '=', '>',
     '?', '@', '^', '_', '`', '{', '|', '}', '~']
)

#: Base 85 digit to integer lookup table.
BASE_85_DICT = dict(_zip(BASE_85, range(0, 86)))


def ipv6_to_base85(addr):
    """Convert a regular IPv6 address to base 85."""
    ip = IPAddress(addr)
    int_val = int(ip)

    remainder = []
    while int_val > 0:
        remainder.append(int_val % 85)
        int_val //= 85

    encoded = ''.join([BASE_85[w] for w in reversed(remainder)])
    leading_zeroes = (20 - len(encoded)) * "0"
    return leading_zeroes + encoded


def base85_to_ipv6(addr):
    """
    Convert a base 85 IPv6 address to its hexadecimal format.
    """
    tokens = list(addr)

    if len(tokens) != 20:
        raise AddrFormatError('Invalid base 85 IPv6 address: %r' % (addr,))

    result = 0
    for i, num in enumerate(reversed(tokens)):
        num = BASE_85_DICT[num]
        result += (num * 85 ** i)

    ip = IPAddress(result, 6)

    return str(ip)



"""
Manage information about files on the minion, set/read user, group
data, modify the ACL of files/directories

:depends:   - win32api
            - win32file
            - win32con
            - salt.utils.win_dacl
"""

import errno
import logging
import os
import os.path
import pathlib
import stat
import sys
import tempfile

import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.user
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.modules.file import (
    __clean_tmp,
    _add_flags,
    _assert_occurrence,
    _binary_replace,
    _check_sig,
    _error,
    _get_bkroot,
    _get_eol,
    _get_flags,
    _mkstemp_copy,
    _regex_to_static,
    _set_line,
    _set_line_eol,
    _set_line_indent,
    _splitlines_preserving_trailing_newline,
    access,
    append,
    apply_template_on_contents,
    basename,
    blockreplace,
    check_file_meta,
    check_hash,
    check_managed,
    check_managed_changes,
    comment,
    comment_line,
    contains,
    contains_glob,
    contains_regex,
    copy,
    delete_backup,
    directory_exists,
    dirname,
    extract_hash,
    file_exists,
    find,
    get_diff,
    get_hash,
    get_managed,
    get_source_sum,
    get_sum,
    join,
    line,
    list_backups,
    list_backups_dir,
    lstat,
    manage_file,
)
from salt.modules.file import normpath as normpath_
from salt.modules.file import (
    pardir,
    patch,
    path_exists_glob,
    prepend,
    psed,
    read,
    readdir,
    readlink,
    rename,
    replace,
    restore_backup,
    rmdir,
    search,
    seek_read,
    seek_write,
    source_list,
    touch,
    truncate,
    uncomment,
    write,
)
from salt.utils.functools import namespaced_function

HAS_WINDOWS_MODULES = False
try:
    if salt.utils.platform.is_windows():
        import pywintypes
        import win32api
        import win32con
        import win32file
        import win32security

        import salt.platform.win

        HAS_WINDOWS_MODULES = True
except ImportError:
    HAS_WINDOWS_MODULES = False

HAS_WIN_DACL = False
try:
    if salt.utils.platform.is_windows():
        import salt.utils.win_dacl

        HAS_WIN_DACL = True
except ImportError:
    HAS_WIN_DACL = False

if salt.utils.platform.is_windows():
    if HAS_WINDOWS_MODULES:
        # namespace functions from file.py
        replace = namespaced_function(replace, globals())
        search = namespaced_function(search, globals())
        _get_flags = namespaced_function(_get_flags, globals())
        _binary_replace = namespaced_function(_binary_replace, globals())
        _check_sig = namespaced_function(_check_sig, globals())
        _splitlines_preserving_trailing_newline = namespaced_function(
            _splitlines_preserving_trailing_newline, globals()
        )
        _error = namespaced_function(_error, globals())
        _get_bkroot = namespaced_function(_get_bkroot, globals())
        list_backups = namespaced_function(list_backups, globals())
        restore_backup = namespaced_function(restore_backup, globals())
        delete_backup = namespaced_function(delete_backup, globals())
        extract_hash = namespaced_function(extract_hash, globals())
        append = namespaced_function(append, globals())
        get_managed = namespaced_function(get_managed, globals())
        check_managed = namespaced_function(check_managed, globals())
        check_managed_changes = namespaced_function(check_managed_changes, globals())
        check_file_meta = namespaced_function(check_file_meta, globals())
        manage_file = namespaced_function(manage_file, globals())
        source_list = namespaced_function(source_list, globals())
        file_exists = namespaced_function(file_exists, globals())
        __clean_tmp = namespaced_function(__clean_tmp, globals())
        directory_exists = namespaced_function(directory_exists, globals())
        touch = namespaced_function(touch, globals())
        contains = namespaced_function(contains, globals())
        contains_regex = namespaced_function(contains_regex, globals())
        contains_glob = namespaced_function(contains_glob, globals())
        get_source_sum = namespaced_function(get_source_sum, globals())
        find = namespaced_function(find, globals())
        psed = namespaced_function(psed, globals())
        get_sum = namespaced_function(get_sum, globals())
        check_hash = namespaced_function(check_hash, globals())
        get_hash = namespaced_function(get_hash, globals())
        get_diff = namespaced_function(get_diff, globals())
        line = namespaced_function(line, globals())
        access = namespaced_function(access, globals())
        copy = namespaced_function(copy, globals())
        readdir = namespaced_function(readdir, globals())
        readlink = namespaced_function(readlink, globals())
        read = namespaced_function(read, globals())
        rmdir = namespaced_function(rmdir, globals())
        truncate = namespaced_function(truncate, globals())
        blockreplace = namespaced_function(blockreplace, globals())
        prepend = namespaced_function(prepend, globals())
        seek_read = namespaced_function(seek_read, globals())
        seek_write = namespaced_function(seek_write, globals())
        rename = namespaced_function(rename, globals())
        lstat = namespaced_function(lstat, globals())
        path_exists_glob = namespaced_function(path_exists_glob, globals())
        write = namespaced_function(write, globals())
        pardir = namespaced_function(pardir, globals())
        join = namespaced_function(join, globals())
        comment = namespaced_function(comment, globals())
        uncomment = namespaced_function(uncomment, globals())
        comment_line = namespaced_function(comment_line, globals())
        _regex_to_static = namespaced_function(_regex_to_static, globals())
        _set_line = namespaced_function(_set_line, globals())
        _set_line_indent = namespaced_function(_set_line_indent, globals())
        _set_line_eol = namespaced_function(_set_line_eol, globals())
        _get_eol = namespaced_function(_get_eol, globals())
        _mkstemp_copy = namespaced_function(_mkstemp_copy, globals())
        _add_flags = namespaced_function(_add_flags, globals())
        apply_template_on_contents = namespaced_function(
            apply_template_on_contents, globals()
        )
        dirname = namespaced_function(dirname, globals())
        basename = namespaced_function(basename, globals())
        list_backups_dir = namespaced_function(list_backups_dir, globals())
        normpath_ = namespaced_function(normpath_, globals())
        _assert_occurrence = namespaced_function(_assert_occurrence, globals())
        patch = namespaced_function(patch, globals())

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "file"


def __virtual__():
    """
    Only works on Windows systems
    """
    if not salt.utils.platform.is_windows() or not HAS_WINDOWS_MODULES:
        return False, "Module win_file: Missing Win32 modules"
    if not HAS_WIN_DACL:
        return False, "Module win_file: Unable to load salt.utils.win_dacl"
    return __virtualname__


__outputter__ = {
    "touch": "txt",
    "append": "txt",
}

__func_alias__ = {
    "makedirs_": "makedirs",
    "normpath_": "normpath",
}


def _resolve_symlink(path, max_depth=64):
    """
    Resolves the given symlink path to its real path, up to a maximum of the
    `max_depth` parameter which defaults to 64.

    If the path is not a symlink path, it is simply returned.
    """
    if sys.getwindowsversion().major < 6:
        raise SaltInvocationError(
            "Symlinks are only supported on Windows Vista or later."
        )

    # make sure we don't get stuck in a symlink loop!
    paths_seen = {path}
    cur_depth = 0
    while is_link(path):
        path = readlink(path)
        if path in paths_seen:
            raise CommandExecutionError("The given path is involved in a symlink loop.")
        paths_seen.add(path)
        cur_depth += 1
        if cur_depth > max_depth:
            raise CommandExecutionError("Too many levels of symbolic links.")

    return path


def gid_to_group(gid):
    """
    Convert the group id to the group name on this system

    Under Windows, because groups are just another ACL entity, this function
    behaves the same as uid_to_user.

    For maintaining Windows systems, this function is superfluous and only
    exists for API compatibility with Unix. Use the uid_to_user function
    instead; an info level log entry will be generated if this function is used
    directly.

    Args:

        gid (str): The gid of the group

    Returns:
        str: The name of the group

    CLI Example:

    .. code-block:: bash

        salt '*' file.gid_to_group 'S-1-5-21-626487655-2533044672-482107328-1010'
    """
    func_name = f"{__virtualname__}.gid_to_group"
    log.info(
        "The function %s should not be used on Windows systems; "
        "see function docs for details.",
        func_name,
    )
    return uid_to_user(gid)


def group_to_gid(group):
    """
    Convert the group to the gid on this system

    Under Windows, because groups are just another ACL entity, this function
    behaves the same as user_to_uid, except if None is given, '' is returned.

    For maintaining Windows systems, this function is superfluous and only
    exists for API compatibility with Unix. Use the user_to_uid function
    instead; an info level log entry will be generated if this function is used
    directly.

    Args:

        group (str): The name of the group

    Returns:
        str: The gid of the group

    CLI Example:

    .. code-block:: bash

        salt '*' file.group_to_gid 'administrators'
    """
    func_name = f"{__virtualname__}.group_to_gid"
    log.info(
        "The function %s should not be used on Windows systems; "
        "see function docs for details.",
        func_name,
    )
    if group is None:
        return ""

    return salt.utils.win_dacl.get_sid_string(group)


def get_pgid(path, follow_symlinks=True):
    """
    Return the id of the primary group that owns a given file (Windows only)

    This function will return the rarely used primary group of a file. This
    generally has no bearing on permissions unless intentionally configured
    and is most commonly used to provide Unix compatibility (e.g. Services
    For Unix, NFS services).

    Ensure you know what you are doing before using this function.

    Args:

        path (str): The path to the file or directory

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        str: The gid of the primary group

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_pgid 'c:\\temp\\test.txt'
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    # Under Windows, if the path is a symlink, the user that owns the symlink is
    # returned, not the user that owns the file/directory the symlink is
    # pointing to. This behavior is *different* to *nix, therefore the symlink
    # is first resolved manually if necessary. Remember symlinks are only
    # supported on Windows Vista or later.
    if follow_symlinks and sys.getwindowsversion().major >= 6:
        path = _resolve_symlink(path)

    group_name = salt.utils.win_dacl.get_primary_group(path)
    return salt.utils.win_dacl.get_sid_string(group_name)


def get_pgroup(path, follow_symlinks=True):
    """
    Return the name of the primary group that owns a given file (Windows only)

    This function will return the rarely used primary group of a file. This
    generally has no bearing on permissions unless intentionally configured
    and is most commonly used to provide Unix compatibility (e.g. Services
    For Unix, NFS services).

    Ensure you know what you are doing before using this function.

    The return value may be 'None', e.g. if the user is not on a domain. This is
    a valid group - do not confuse this with the Salt/Python value of None which
    means no value was returned. To be certain, use the `get_pgid` function
    which will return the SID, including for the system 'None' group.

    Args:

        path (str): The path to the file or directory

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        str: The name of the primary group

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_pgroup 'c:\\temp\\test.txt'
    """
    return uid_to_user(get_pgid(path, follow_symlinks))


def get_gid(path, follow_symlinks=True):
    """
    Return the id of the group that owns a given file

    Under Windows, this will return the uid of the file.

    While a file in Windows does have a 'primary group', this rarely used
    attribute generally has no bearing on permissions unless intentionally
    configured and is only used to support Unix compatibility features (e.g.
    Services For Unix, NFS services).

    Salt, therefore, remaps this function to provide functionality that
    somewhat resembles Unix behavior for API compatibility reasons. When
    managing Windows systems, this function is superfluous and will generate
    an info level log entry if used directly.

    If you do actually want to access the 'primary group' of a file, use
    `file.get_pgid`.

    Args:

        path (str): The path to the file or directory

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        str: The gid of the owner

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_gid 'c:\\temp\\test.txt'
    """
    func_name = f"{__virtualname__}.get_gid"
    log.info(
        "The function %s should not be used on Windows systems; "
        "see function docs for details. The value returned is the "
        "uid.",
        func_name,
    )
    return get_uid(path, follow_symlinks)


def get_group(path, follow_symlinks=True):
    """
    Return the group that owns a given file

    Under Windows, this will return the user (owner) of the file.

    While a file in Windows does have a 'primary group', this rarely used
    attribute generally has no bearing on permissions unless intentionally
    configured and is only used to support Unix compatibility features (e.g.
    Services For Unix, NFS services).

    Salt, therefore, remaps this function to provide functionality that
    somewhat resembles Unix behavior for API compatibility reasons. When
    managing Windows systems, this function is superfluous and will generate
    an info level log entry if used directly.

    If you do actually want to access the 'primary group' of a file, use
    `file.get_pgroup`.

    Args:

        path (str): The path to the file or directory

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        str: The name of the owner

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_group 'c:\\temp\\test.txt'
    """
    func_name = f"{__virtualname__}.get_group"
    log.info(
        "The function %s should not be used on Windows systems; "
        "see function docs for details. The value returned is the "
        "user (owner).",
        func_name,
    )
    return get_user(path, follow_symlinks)


def uid_to_user(uid):
    """
    Convert a User ID (uid) to a username

    Args:

        uid (str): The user id to lookup

    Returns:
        str: The name of the user. The ``uid`` will be returned if there is no
             corresponding username

    CLI Example:

    .. code-block:: bash

        salt '*' file.uid_to_user 'S-1-5-21-626487655-2533044672-482107328-1010'
    """
    if uid is None or uid == "":
        return ""

    return salt.utils.win_dacl.get_name(uid)


def user_to_uid(user):
    """
    Convert user name to a uid

    Args:

        user (str): The user to lookup

    Returns:
        str: The user id of the user

    CLI Example:

    .. code-block:: bash

        salt '*' file.user_to_uid 'myusername'
    """
    if user is None:
        user = salt.utils.user.get_user()

    return salt.utils.win_dacl.get_sid_string(user)


def get_uid(path, follow_symlinks=True):
    """
    Return the id of the user that owns a given file

    Symlinks are followed by default to mimic Unix behavior. Specify
    `follow_symlinks=False` to turn off this behavior.

    Args:

        path (str): The path to the file or directory

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        str: The uid of the owner

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_uid 'c:\\temp\\test.txt'
        salt '*' file.get_uid 'c:\\temp\\test.txt' follow_symlinks=False
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    # Under Windows, if the path is a symlink, the user that owns the symlink is
    # returned, not the user that owns the file/directory the symlink is
    # pointing to. This behavior is *different* to *nix, therefore the symlink
    # is first resolved manually if necessary. Remember symlinks are only
    # supported on Windows Vista or later.
    if follow_symlinks and sys.getwindowsversion().major >= 6:
        path = _resolve_symlink(path)

    owner_sid = salt.utils.win_dacl.get_owner(path)
    return salt.utils.win_dacl.get_sid_string(owner_sid)


def get_user(path, follow_symlinks=True):
    """
    Return the user that owns a given file

    Symlinks are followed by default to mimic Unix behavior. Specify
    `follow_symlinks=False` to turn off this behavior.

    Args:

        path (str): The path to the file or directory

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        str: The name of the owner

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_user 'c:\\temp\\test.txt'
        salt '*' file.get_user 'c:\\temp\\test.txt' follow_symlinks=False
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    # Under Windows, if the path is a symlink, the user that owns the symlink is
    # returned, not the user that owns the file/directory the symlink is
    # pointing to. This behavior is *different* to *nix, therefore the symlink
    # is first resolved manually if necessary. Remember symlinks are only
    # supported on Windows Vista or later.
    if follow_symlinks and sys.getwindowsversion().major >= 6:
        path = _resolve_symlink(path)

    return salt.utils.win_dacl.get_owner(path)


def get_mode(path):
    """
    Return the mode of a file

    Right now we're just returning None because Windows' doesn't have a mode
    like Linux

    Args:

        path (str): The path to the file or directory

    Returns:
        None

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_mode /etc/passwd
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    func_name = f"{__virtualname__}.get_mode"
    log.info(
        "The function %s should not be used on Windows systems; "
        "see function docs for details. The value returned is "
        "always None.",
        func_name,
    )
    return None


def lchown(path, user, group=None, pgroup=None):
    """
    Chown a file, pass the file the desired user and group without following any
    symlinks.

    Under Windows, the group parameter will be ignored.

    This is because while files in Windows do have a 'primary group'
    property, this is rarely used.  It generally has no bearing on
    permissions unless intentionally configured and is most commonly used to
    provide Unix compatibility (e.g. Services For Unix, NFS services).

    If you do want to change the 'primary group' property and understand the
    implications, pass the Windows only parameter, pgroup, instead.

    To set the primary group to 'None', it must be specified in quotes.
    Otherwise, Salt will interpret it as the Python value of None and no primary
    group changes will occur. See the example below.

    Args:

        path (str): The path to the file or directory

        user (str): The name of the user to own the file

        group (:obj:`str`, optional):
            The group (not used).
            Default is ``None``.

        pgroup (:obj:`str`, optional):
            The primary group to assign.
            Default is ``None``.

    Returns:
        bool: True if successful, otherwise error

    CLI Example:

    .. code-block:: bash

        salt '*' file.lchown 'c:\\temp\\test.txt' myusername
        salt '*' file.lchown 'c:\\temp\\test.txt' myusername pgroup=Administrators
        salt '*' file.lchown 'c:\\temp\\test.txt' myusername "pgroup='None'"
    """
    if group:
        func_name = f"{__virtualname__}.lchown"
        log.info(
            "The group parameter has no effect when using %s on "
            "Windows systems; see function docs for details.",
            func_name,
        )
        log.debug("win_file.py %s Ignoring the group parameter for %s", func_name, path)
        group = None

    return chown(path, user, group, pgroup, follow_symlinks=False)


def chown(path, user, group=None, pgroup=None, follow_symlinks=True):
    """
    Chown a file, pass the file the desired user and group

    Under Windows, the group parameter will be ignored.

    This is because while files in Windows do have a 'primary group'
    property, this is rarely used.  It generally has no bearing on
    permissions unless intentionally configured and is most commonly used to
    provide Unix compatibility (e.g. Services For Unix, NFS services).

    If you do want to change the 'primary group' property and understand the
    implications, pass the Windows only parameter, pgroup, instead.

    Args:

        path (str): The path to the file or directory

        user (str): The name of the user to own the file

        group (:obj:`str`, optional):
            The group (not used).
            Default is ``None``.

        pgroup (:obj:`str`, optional):
            The primary group to assign. Default is ``None``.

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: Path not found
        CommandExecutionError: Failed to change ownership

    CLI Example:

    .. code-block:: bash

        salt '*' file.chown 'c:\\temp\\test.txt' 'myusername'
        salt '*' file.chown 'c:\\temp\\test.txt' 'myusername' pgroup=Administrators
        salt '*' file.chown 'c:\\temp\\test.txt' 'myusername' pgroup=None
    """
    # the group parameter is not used; only provided for API compatibility
    if group is not None:
        func_name = f"{__virtualname__}.chown"
        log.info(
            "The group parameter has no effect when using %s on "
            "Windows systems; see function docs for details.",
            func_name,
        )
        log.debug("win_file.py %s Ignoring the group parameter for %s", func_name, path)

    if follow_symlinks and sys.getwindowsversion().major >= 6:
        path = _resolve_symlink(path)

    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    salt.utils.win_dacl.set_owner(path, user)
    if pgroup:
        salt.utils.win_dacl.set_primary_group(path, pgroup)

    return True


def chpgrp(path, group):
    """
    Change the group of a file

    Under Windows, this will set the rarely used primary group of a file.
    This generally has no bearing on permissions unless intentionally
    configured and is most commonly used to provide Unix compatibility (e.g.
    Services For Unix, NFS services).

    Ensure you know what you are doing before using this function.

    Args:

        path (str): The path to the file or directory

        pgroup (str): The primary group to assign

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: Failed to change group

    CLI Example:

    .. code-block:: bash

        salt '*' file.chpgrp 'c:\\temp\\test.txt' Administrators
        salt '*' file.chpgrp 'c:\\temp\\test.txt' None
    """
    return salt.utils.win_dacl.set_primary_group(path, group)


def chgrp(path, group):
    """
    Change the group of a file

    Under Windows, this will do nothing.

    While a file in Windows does have a 'primary group', this rarely used
    attribute generally has no bearing on permissions unless intentionally
    configured and is only used to support Unix compatibility features (e.g.
    Services For Unix, NFS services).

    Salt, therefore, remaps this function to do nothing while still being
    compatible with Unix behavior. When managing Windows systems, this function
    is superfluous and will generate an info level log entry if used directly.

    If you do actually want to set the 'primary group' of a file, use
    ``file.chpgrp``.

    To set group permissions use ``file.set_perms``.

    Args:

        path (str): The path to the file or directory

        group (str): The group (unused)

    Returns:
        None

    CLI Example:

    .. code-block:: bash

        salt '*' file.chgrp 'c:\\temp\\test.txt' administrators
    """
    func_name = f"{__virtualname__}.chgrp"
    log.info(
        "The function %s should not be used on Windows systems; see "
        "function docs for details.",
        func_name,
    )
    log.debug("win_file.py %s Doing nothing for %s", func_name, path)

    return None


def stats(path, hash_type="sha256", follow_symlinks=True):
    """
    Return a dict containing the stats about a given file

    Under Windows, `gid` will equal `uid` and `group` will equal `user`.

    While a file in Windows does have a 'primary group', this rarely used
    attribute generally has no bearing on permissions unless intentionally
    configured and is only used to support Unix compatibility features (e.g.
    Services For Unix, NFS services).

    Salt, therefore, remaps these properties to keep some kind of compatibility
    with Unix behavior. If the 'primary group' is required, it can be accessed
    in the `pgroup` and `pgid` properties.

    Args:

        path (str): The path to the file or directory

        hash_type (:obj:`str`, optional):
            The type of hash to return.
            Default is``sha256``.

        follow_symlinks (:obj:`bool`, optional):
            If the object specified by ``path`` is a symlink, get attributes of
            the linked file instead of the symlink itself.
            Default is ``True``.

    Returns:
        dict: A dictionary of file/directory stats

    CLI Example:

    .. code-block:: bash

        salt '*' file.stats /etc/passwd
    """
    # This is to mirror the behavior of file.py. `check_file_meta` expects an
    # empty dictionary when the file does not exist
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    if follow_symlinks and sys.getwindowsversion().major >= 6:
        path = _resolve_symlink(path)

    pstat = os.stat(path)

    ret = {}
    ret["inode"] = pstat.st_ino
    # don't need to resolve symlinks again because we've already done that
    ret["uid"] = get_uid(path, follow_symlinks=False)
    # maintain the illusion that group is the same as user as states need this
    ret["gid"] = ret["uid"]
    ret["user"] = uid_to_user(ret["uid"])
    ret["group"] = ret["user"]
    ret["pgid"] = get_pgid(path, follow_symlinks)
    ret["pgroup"] = gid_to_group(ret["pgid"])
    ret["atime"] = pstat.st_atime
    ret["mtime"] = pstat.st_mtime
    ret["ctime"] = pstat.st_ctime
    ret["size"] = pstat.st_size
    ret["mode"] = salt.utils.files.normalize_mode(oct(stat.S_IMODE(pstat.st_mode)))
    if hash_type:
        ret["sum"] = get_sum(path, hash_type)
    ret["type"] = "file"
    if stat.S_ISDIR(pstat.st_mode):
        ret["type"] = "dir"
    if stat.S_ISCHR(pstat.st_mode):
        ret["type"] = "char"
    if stat.S_ISBLK(pstat.st_mode):
        ret["type"] = "block"
    if stat.S_ISREG(pstat.st_mode):
        ret["type"] = "file"
    if stat.S_ISLNK(pstat.st_mode):
        ret["type"] = "link"
    if stat.S_ISFIFO(pstat.st_mode):
        ret["type"] = "pipe"
    if stat.S_ISSOCK(pstat.st_mode):
        ret["type"] = "socket"
    ret["target"] = os.path.realpath(path)
    return ret


def _get_version_os(flags):
    """
    Helper function to parse the OS data

    Args:

        flags: The flags as returned by the GetFileVersionInfo function

    Returns:
        list: A list of Operating system properties found in the flag
    """
    file_os = []
    file_os_flags = {
        0x00000001: "16-bit Windows",
        0x00000002: "16-bit Presentation Manager",
        0x00000003: "32-bit Presentation Manager",
        0x00000004: "32-bit Windows",
        0x00010000: "MS-DOS",
        0x00020000: "16-bit OS/2",
        0x00030000: "32-bit OS/2",
        0x00040000: "Windows NT",
    }
    for item in file_os_flags:
        if item & flags == item:
            file_os.append(file_os_flags[item])
    return file_os


def _get_version_type(file_type, file_subtype):
    ret_type = None
    file_types = {
        0x00000001: "Application",
        0x00000002: "DLL",
        0x00000003: "Driver",
        0x00000004: "Font",
        0x00000005: "Virtual Device",
        0x00000007: "Static Link Library",
    }
    driver_subtypes = {
        0x00000001: "Printer",
        0x00000002: "Keyboard",
        0x00000003: "Language",
        0x00000004: "Display",
        0x00000005: "Mouse",
        0x00000006: "Network",
        0x00000007: "System",
        0x00000008: "Installable",
        0x00000009: "Sound",
        0x0000000A: "Communications",
        0x0000000C: "Versioned Printer",
    }
    font_subtypes = {
        0x00000001: "Raster",
        0x00000002: "Vector",
        0x00000003: "TrueType",
    }
    if file_type in file_types:
        ret_type = file_types[file_type]

    if ret_type == "Driver":
        if file_subtype in driver_subtypes:
            ret_type = f"{driver_subtypes[file_subtype]} Driver"
    if ret_type == "Font":
        if file_subtype in font_subtypes:
            ret_type = f"{font_subtypes[file_subtype]} Font"
    if ret_type == "Virtual Device":
        # The Virtual Device Identifier
        ret_type = f"Virtual Device: {file_subtype}"
    return ret_type


def _get_version(path, fixed_info=None):
    """
    Get the version of the file passed in path or the fixed_info object if
    passed.

    Args:

        path (str): The path to the file

        fixed_info (:obj:`obj`, optional):
            The fixed info object returned by the GetFileVersionInfo function.
            Default is ``None``.

    Returns:
        str: The version of the file
    """
    if not fixed_info:
        try:
            # Backslash returns a VS_FIXEDFILEINFO structure
            # https://docs.microsoft.com/en-us/windows/win32/api/verrsrc/ns-verrsrc-vs_fixedfileinfo
            fixed_info = win32api.GetFileVersionInfo(path, "\\")
        except pywintypes.error:
            log.debug("No version info found: %s", path)
            return ""

    return "{}.{}.{}.{}".format(
        win32api.HIWORD(fixed_info["FileVersionMS"]),
        win32api.LOWORD(fixed_info["FileVersionMS"]),
        win32api.HIWORD(fixed_info["FileVersionLS"]),
        win32api.LOWORD(fixed_info["FileVersionLS"]),
    )


def version(path):
    r"""
    .. versionadded:: 3005

    Get the version of a file.

    .. note::
        Not all files have version information. The following are common file
        types that contain version information:

            - .exe
            - .dll
            - .sys

    Args:

        path (str): The path to the file.

    Returns:
        str: The version of the file if the file contains it. Otherwise, an
            empty string will be returned.

    Raises:
        CommandExecutionError: If the file does not exist
        CommandExecutionError: If the path is not a file

    CLI Example:

    .. code-block:: bash

        salt '*' file.version 'C:\Windows\notepad.exe'
    """
    # Input validation
    if not os.path.exists(path):
        raise CommandExecutionError(f"File not found: {path}")
    if os.path.isdir(path):
        raise CommandExecutionError(f"Not a file: {path}")
    return _get_version(path)


def version_details(path):
    r"""
    .. versionadded:: 3005

    Get file details for a file. Similar to what's in the details tab on the
    file properties.

    .. note::
        Not all files have version information. The following are common file
        types that contain version information:

            - .exe
            - .dll
            - .sys

    Args:
        path (str): The path to the file.

    Returns:
        dict: A dictionary containing details about the file related to version.
            An empty dictionary if the file contains no version information.

    Raises:
        CommandExecutionError: If the file does not exist
        CommandExecutionError: If the path is not a file

    CLI Example:

    .. code-block:: bash

        salt '*' file.version_details 'C:\Windows\notepad.exe'
    """
    # Input validation
    if not os.path.exists(path):
        raise CommandExecutionError(f"File not found: {path}")
    if os.path.isdir(path):
        raise CommandExecutionError(f"Not a file: {path}")

    ret = {}
    try:
        # Backslash returns a VS_FIXEDFILEINFO structure
        # https://docs.microsoft.com/en-us/windows/win32/api/verrsrc/ns-verrsrc-vs_fixedfileinfo
        fixed_info = win32api.GetFileVersionInfo(path, "\\")
    except pywintypes.error:
        log.debug("No version info found: %s", path)
        return ret

    ret["Version"] = _get_version(path, fixed_info)
    ret["OperatingSystem"] = _get_version_os(fixed_info["FileOS"])
    ret["FileType"] = _get_version_type(
        fixed_info["FileType"], fixed_info["FileSubtype"]
    )

    try:
        # \VarFileInfo\Translation returns a list of available
        # (language, codepage) pairs that can be used to retrieve string info.
        # We only care about the first pair.
        # https://docs.microsoft.com/en-us/windows/win32/menurc/varfileinfo-block
        language, codepage = win32api.GetFileVersionInfo(
            path, "\\VarFileInfo\\Translation"
        )[0]
    except pywintypes.error:
        log.debug("No extended version info found: %s", path)
        return ret

    # All other properties are in the StringFileInfo block
    # \StringFileInfo\<hex language><hex codepage>\<property name>
    # https://docs.microsoft.com/en-us/windows/win32/menurc/stringfileinfo-block
    property_names = (
        "Comments",
        "CompanyName",
        "FileDescription",
        "FileVersion",
        "InternalName",
        "LegalCopyright",
        "LegalTrademarks",
        "OriginalFilename",
        "PrivateBuild",
        "ProductName",
        "ProductVersion",
        "SpecialBuild",
    )
    for prop_name in property_names:
        str_info_path = "\\StringFileInfo\\{:04X}{:04X}\\{}".format(
            language, codepage, prop_name
        )
        try:
            ret[prop_name] = win32api.GetFileVersionInfo(path, str_info_path)
        except pywintypes.error:
            pass

    return ret


def get_attributes(path):
    """
    Return a dictionary object with the Windows file attributes for a file.

    Args:

        path (str): The path to the file or directory

    Returns:
        dict: A dictionary of file attributes

    CLI Example:

    .. code-block:: bash

        salt '*' file.get_attributes 'c:\\temp\\a.txt'
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    # set up dictionary for attribute values
    attributes = {}

    # Get cumulative int value of attributes
    intAttributes = win32file.GetFileAttributes(path)

    # Assign individual attributes
    attributes["archive"] = (intAttributes & 32) == 32
    attributes["reparsePoint"] = (intAttributes & 1024) == 1024
    attributes["compressed"] = (intAttributes & 2048) == 2048
    attributes["directory"] = (intAttributes & 16) == 16
    attributes["encrypted"] = (intAttributes & 16384) == 16384
    attributes["hidden"] = (intAttributes & 2) == 2
    attributes["normal"] = (intAttributes & 128) == 128
    attributes["notIndexed"] = (intAttributes & 8192) == 8192
    attributes["offline"] = (intAttributes & 4096) == 4096
    attributes["readonly"] = (intAttributes & 1) == 1
    attributes["system"] = (intAttributes & 4) == 4
    attributes["temporary"] = (intAttributes & 256) == 256

    # check if it's a Mounted Volume
    attributes["mountedVolume"] = False
    if attributes["reparsePoint"] is True and attributes["directory"] is True:
        fileIterator = win32file.FindFilesIterator(path)
        findDataTuple = next(fileIterator)
        if findDataTuple[6] == 0xA0000003:
            attributes["mountedVolume"] = True
    # check if it's a soft (symbolic) link

    # Note:  os.path.islink() does not work in
    #   Python 2.7 for the Windows NTFS file system.
    #   The following code does, however, work (tested in Windows 8)

    attributes["symbolicLink"] = False
    if attributes["reparsePoint"] is True:
        fileIterator = win32file.FindFilesIterator(path)
        findDataTuple = next(fileIterator)
        if findDataTuple[6] == 0xA000000C:
            attributes["symbolicLink"] = True

    return attributes


def set_attributes(
    path,
    archive=None,
    hidden=None,
    normal=None,
    notIndexed=None,
    readonly=None,
    system=None,
    temporary=None,
):
    """
    Set file attributes for a file. Note that the normal attribute means that
    all others are false. So setting it will clear all others.

    Args:

        path (str): The path to the file or directory

        archive (:obj:`bool`, optional):
            Sets the archive attribute.
            Default is ``None``.

        hidden (:obj:`bool`, optional):
            Sets the hidden attribute.
            Default is ``None``.

        normal (:obj:`bool`, optional):
            Resets the file attributes. Cannot be used in conjunction with any
            other attribute.
            Default is ``None``.

        notIndexed (:obj:`bool`, optional):
            Sets the indexed attribute.
            Default is ``None``.

        readonly (:obj:`bool`, optional):
            Sets the readonly attribute.
            Default is ``None``.

        system (:obj:`bool`, optional):
            Sets the system attribute.
            Default is ``None``.

        temporary (:obj:`bool`, optional):
            Sets the temporary attribute.
            Default is ``None``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' file.set_attributes 'c:\\temp\\a.txt' normal=True
        salt '*' file.set_attributes 'c:\\temp\\a.txt' readonly=True hidden=True
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    if normal:
        if archive or hidden or notIndexed or readonly or system or temporary:
            raise CommandExecutionError(
                "Normal attribute may not be used with any other attributes"
            )
        ret = win32file.SetFileAttributes(path, 128)
        return True if ret is None else False

    # Get current attributes
    intAttributes = win32file.GetFileAttributes(path)

    # individually set or clear bits for appropriate attributes
    if archive is not None:
        if archive:
            intAttributes |= 0x20
        else:
            intAttributes &= 0xFFDF
    if hidden is not None:
        if hidden:
            intAttributes |= 0x2
        else:
            intAttributes &= 0xFFFD
    if notIndexed is not None:
        if notIndexed:
            intAttributes |= 0x2000
        else:
            intAttributes &= 0xDFFF
    if readonly is not None:
        if readonly:
            intAttributes |= 0x1
        else:
            intAttributes &= 0xFFFE
    if system is not None:
        if system:
            intAttributes |= 0x4
        else:
            intAttributes &= 0xFFFB
    if temporary is not None:
        if temporary:
            intAttributes |= 0x100
        else:
            intAttributes &= 0xFEFF

    ret = win32file.SetFileAttributes(path, intAttributes)
    return True if ret is None else False


def set_mode(path, mode):
    """
    Set the mode of a file

    This just calls get_mode, which returns None because we don't use mode on
    Windows

    Args:

        path (str): The path to the file or directory

        mode (str): The mode (not used)

    Returns:
        None

    CLI Example:

    .. code-block:: bash

        salt '*' file.set_mode /etc/passwd 0644
    """
    func_name = f"{__virtualname__}.set_mode"
    log.info(
        "The function %s should not be used on Windows systems; "
        "see function docs for details. The value returned is "
        "always None. Use set_perms instead.",
        func_name,
    )
    return get_mode(path)


def remove(path, force=False):
    """
    Remove the named file or directory

    Args:

        path (str): The path to the file or directory to remove.

        force (:obj:`bool`, optional):
            Remove even if marked Read-Only.
            Default is ``False``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' file.remove 'C:\\Temp'
    """
    # This must be a recursive function in windows to properly deal with
    # Symlinks. The shutil.rmtree function will remove the contents of
    # the Symlink source in windows.

    path = pathlib.Path(os.path.expanduser(path))

    if not path.is_absolute():
        raise SaltInvocationError(f"File path must be absolute: {path}")

    # Does the file/folder exists
    if not path.exists() and not path.is_symlink():
        raise CommandExecutionError(f"Path not found: {path}")

    # Remove ReadOnly Attribute
    file_attributes = win32api.GetFileAttributes(str(path))
    if force:
        # Get current file attributes
        win32api.SetFileAttributes(str(path), win32con.FILE_ATTRIBUTE_NORMAL)

    try:
        if path.is_file() or path.is_symlink():
            # A file and a symlinked file are removed the same way
            path.unlink()
        else:
            # Twangboy: This is for troubleshooting
            is_dir = os.path.isdir(path)
            exists = os.path.exists(path)
            # This is a directory, list its contents and remove them recursively
            for child in path.iterdir():
                # If it's a normal directory, recurse to remove its contents
                remove(str(child), force)
            # rmdir will work now because the directory is empty
            path.rmdir()
    except OSError as exc:
        if force:
            # Reset attributes to the original if delete fails.
            win32api.SetFileAttributes(str(path), file_attributes)
        raise CommandExecutionError(f"Could not remove '{path}': {exc}")

    return True


def symlink(src, link, force=False, atomic=False, follow_symlinks=True):
    """
    Create a symbolic link to a file

    This is only supported with Windows Vista or later and must be executed by
    a user with the SeCreateSymbolicLink privilege.

    The behavior of this function matches the Unix equivalent, with one
    exception - invalid symlinks cannot be created. The source path must exist.
    If it doesn't, an error will be raised.

    Args:

        src (str): The path to a file or directory

        link (str): The path to the link. Must be an absolute path

        force (:obj:`bool`, optional):
            Overwrite an existing symlink with the same name.
            Default is ``False``.

            .. versionadded:: 3005

        atomic (:obj:`bool`, optional):
            Use atomic file operations to create the symlink.
            Default is ``False``.

            .. versionadded:: 3006.0

        follow_symlinks (bool):
            If set to ``False``, use ``os.path.lexists()`` for existence checks
            instead of ``os.path.exists()``.
            .. versionadded:: 3007.0

    Returns:
        bool: ``True`` if successful

    Raises:
        SaltInvocationError: Unsupported version of Windows
        SaltInvocationError: Link path must be absolute
        CommandExecutionError: Existing symlink and force is False
        CommandExecutionError: Existing path is not a symlink
        CommandExecutionError: Failed to create symlink

    CLI Example:

    .. code-block:: bash

        salt '*' file.symlink /path/to/file /path/to/link
    """
    # When Python 3.2 or later becomes the minimum version, this function can be
    # replaced with the built-in os.symlink function, which supports Windows.
    if sys.getwindowsversion().major < 6:
        raise SaltInvocationError(
            "Symlinks are only supported on Windows Vista or later."
        )

    if not os.path.isabs(link):
        raise SaltInvocationError(f"Link path must be absolute: {link}")

    if follow_symlinks:
        exists = os.path.exists
    else:
        exists = os.path.lexists

    if os.path.islink(link):
        try:
            if os.path.normpath(salt.utils.path.readlink(link)) == os.path.normpath(
                src
            ):
                log.debug("link already in correct state: %s -> %s", link, src)
                return True
        except OSError:
            pass

        if not force and not atomic:
            msg = f"Found existing symlink: {link}"
            raise CommandExecutionError(msg)

    if exists(link) and not force and not atomic:
        msg = f"Existing path is not a symlink: {link}"
        raise CommandExecutionError(msg)

    # ensure paths are using the right slashes
    src = os.path.normpath(src)
    link = os.path.normpath(link)

    is_dir = os.path.isdir(src)

    # Elevate the token from the current process
    desired_access = win32security.TOKEN_QUERY | win32security.TOKEN_ADJUST_PRIVILEGES
    th = win32security.OpenProcessToken(win32api.GetCurrentProcess(), desired_access)
    salt.platform.win.elevate_token(th)

    if (os.path.islink(link) or exists(link)) and force and not atomic:
        os.unlink(link)
    elif atomic:
        link_dir = os.path.dirname(link)
        retry = 0
        while retry < 5:
            temp_link = tempfile.mktemp(dir=link_dir)
            try:
                win32file.CreateSymbolicLink(temp_link, src, int(is_dir))
                break
            except win32file.error:
                retry += 1
        try:
            win32file.MoveFileEx(
                temp_link,
                link,
                win32file.MOVEFILE_REPLACE_EXISTING | win32file.MOVEFILE_WRITE_THROUGH,
            )
            return True
        except win32file.error:
            os.remove(temp_link)
            raise CommandExecutionError(f"Could not create '{link}'")

    try:
        win32file.CreateSymbolicLink(link, src, int(is_dir))
        return True
    except win32file.error as exc:
        raise CommandExecutionError(
            f"Could not create '{link}' - [{exc.winerror}] {exc.strerror}"
        )


def is_link(path):
    """
    Check if the path is a symlink

    This is only supported on Windows Vista or later.

    Inline with Unix behavior, this function will raise an error if the path
    is not a symlink, however, the error raised will be a SaltInvocationError,
    not an OSError.

    Args:

        path (str): The path to a file or directory

    Returns:
        bool: ``True`` if path is a symlink, otherwise ``False``

    CLI Example:

    .. code-block:: bash

       salt '*' file.is_link /path/to/link
    """
    if sys.getwindowsversion().major < 6:
        raise SaltInvocationError(
            "Symlinks are only supported on Windows Vista or later."
        )

    try:
        return salt.utils.path.islink(path)
    except Exception as exc:  # pylint: disable=broad-except
        raise CommandExecutionError(exc)


def mkdir(
    path, owner=None, grant_perms=None, deny_perms=None, inheritance=True, reset=False
):
    """
    Ensure that the directory is available and permissions are set.

    Args:

        path (str):
            The full path to the directory.

        owner (:obj:`str`, optional):
            The owner of the directory. If not passed, it will be the account
            that created the directory, likely SYSTEM.
            Default is ``None``.

        grant_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and the basic permissions to
            grant, ie: ``{'user': {'perms': 'basic_permission'}}``. You can also
            set the ``applies_to`` setting here. The default is
            ``this_folder_subfolders_files``. Specify another ``applies_to``
            setting like this:

            .. code-block:: yaml

                {'user': {'perms': 'full_control', 'applies_to': 'this_folder'}}

            To set advanced permissions use a list for the ``perms`` parameter,
            ie:

            .. code-block:: yaml

                {'user': {'perms': ['read_attributes', 'read_ea'], 'applies_to': 'this_folder'}}

            Default is ``None``.

        deny_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and permissions to deny along
            with the ``applies_to`` setting. Use the same format used for the
            ``grant_perms`` parameter. Remember, deny permissions supersede
            grant permissions.
            Default is ``None``.

        inheritance (:obj:`bool`, optional):
            If ``True``, the object will inherit permissions from the parent. If
            ``False``, inheritance will be disabled. Inheritance setting will
            not apply to parent directories if they must be created.
            Default is ``True``.

        reset (:obj:`bool`, optional):
            If ``True`` the existing DACL will be cleared and replaced with the
            settings defined in this function. If ``False``, new entries will be
            appended to the existing DACL.
            Default is ``False``.

            .. versionadded:: 2018.3.0

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: Failed to create directory

    CLI Example:

    .. code-block:: bash

        # To grant the 'Users' group 'read & execute' permissions.
        salt '*' file.mkdir 'C:\\Temp\\' Administrators '{"Users": {"perms": "read_execute"}}'

        # Locally using salt call
        salt-call file.mkdir 'C:\\Temp\\' Administrators '{"Users": {"perms": "read_execute", "applies_to": "this_folder_only"}}'

        # Specify advanced attributes with a list
        salt '*' file.mkdir 'C:\\Temp\\' Administrators '{"jsnuffy": {"perms": ["read_attributes", "read_ea"], "applies_to": "this_folder_only"}}'
    """
    # Make sure the drive is valid
    drive = os.path.splitdrive(path)[0]
    if not os.path.isdir(drive):
        raise CommandExecutionError(f"Drive {drive} is not mapped")

    path = os.path.expanduser(path)
    path = os.path.expandvars(path)

    if not os.path.isdir(path):

        try:
            # Make the directory
            os.mkdir(path)

            # Set owner
            if owner:
                salt.utils.win_dacl.set_owner(obj_name=path, principal=owner)

            # Set permissions
            salt.utils.win_dacl.set_perms(
                obj_name=path,
                obj_type="file",
                grant_perms=grant_perms,
                deny_perms=deny_perms,
                inheritance=inheritance,
                reset=reset,
            )

        except OSError as exc:
            raise CommandExecutionError(exc)

    return True


def makedirs_(
    path, owner=None, grant_perms=None, deny_perms=None, inheritance=True, reset=False
):
    """
    Ensure that the parent directory containing this path is available.

    Args:

        path (str):

            The full path to the directory.

            .. note::

                The path must end with a trailing slash otherwise the
                directory(s) will be created up to the parent directory. For
                example if path is ``C:\\temp\\test``, then it would be treated
                as ``C:\\temp\\`` but if the path ends with a trailing slash
                like ``C:\\temp\\test\\``, then it would be treated as
                ``C:\\temp\\test\\``.

        owner (:obj:`str`, optional):
            The owner of the directory. If not passed, it will be the account
            that created the directory, likely SYSTEM. Default is ``None``.

        grant_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and the basic permissions to
            grant, ie: ``{'user': {'perms': 'basic_permission'}}``. You can also
            set the ``applies_to`` setting here. The default is
            ``this_folder_subfolders_files``. Specify another ``applies_to``
            setting like this:

            .. code-block:: yaml

                {'user': {'perms': 'full_control', 'applies_to': 'this_folder'}}

            To set advanced permissions use a list for the ``perms`` parameter,
            ie:

            .. code-block:: yaml

                {'user': {'perms': ['read_attributes', 'read_ea'], 'applies_to': 'this_folder'}}

            Default is ``None``.

        deny_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and permissions to deny along
            with the ``applies_to`` setting. Use the same format used for the
            ``grant_perms`` parameter. Remember, deny permissions supersede
            grant permissions.
            Default is ``None``.

        inheritance (:obj:`bool`, optional):
            If ``True``, the object will inherit permissions from the parent. If
            ``False``, inheritance will be disabled. Inheritance setting will
            not apply to parent directories if they must be created.
            Default is ``True``.

        reset (:obj:`bool`, optional):
            If ``True``, the existing DACL will be cleared and replaced with the
            settings defined in this function. If ``False``, new entries will be
            appended to the existing DACL.
            Default is ``False``.

            .. versionadded:: 2018.3.0

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: Failed to create directories

    CLI Example:

    .. code-block:: bash

        # To grant the 'Users' group 'read & execute' permissions.
        salt '*' file.makedirs 'C:\\Temp\\' Administrators '{"Users": {"perms": "read_execute"}}'

        # Locally using salt call
        salt-call file.makedirs 'C:\\Temp\\' Administrators '{"Users": {"perms": "read_execute", "applies_to": "this_folder_only"}}'

        # Specify advanced attributes with a list
        salt '*' file.makedirs 'C:\\Temp\\' Administrators '{"jsnuffy": {"perms": ["read_attributes", "read_ea"], "applies_to": "this_folder_only"}}'
    """
    path = os.path.expanduser(path)

    # walk up the directory structure until we find the first existing
    # directory
    dirname = os.path.normpath(os.path.dirname(path))

    if os.path.isdir(dirname):
        # There's nothing for us to do
        msg = f"Directory '{dirname}' already exists"
        log.debug(msg)
        return msg

    if os.path.exists(dirname):
        msg = f"The path '{dirname}' already exists and is not a directory"
        log.debug(msg)
        return msg

    directories_to_create = []
    while True:
        if os.path.isdir(dirname):
            break

        directories_to_create.append(dirname)
        current_dirname = dirname
        dirname = os.path.dirname(dirname)

        if current_dirname == dirname:
            raise SaltInvocationError(
                "Recursive creation for path '{}' would result in an "
                "infinite loop. Please use an absolute path.".format(dirname)
            )

    # create parent directories from the topmost to the most deeply nested one
    directories_to_create.reverse()
    for directory_to_create in directories_to_create:
        # all directories have the user, group and mode set!!
        log.debug("Creating directory: %s", directory_to_create)
        mkdir(
            path=directory_to_create,
            owner=owner,
            grant_perms=grant_perms,
            deny_perms=deny_perms,
            inheritance=inheritance,
            reset=reset,
        )

    return True


def makedirs_perms(
    path, owner=None, grant_perms=None, deny_perms=None, inheritance=True, reset=True
):
    """
    Set owner and permissions for each directory created.

    Args:

        path (str):
            The full path to the directory.

        owner (:obj:`str`, optional):
            The owner of the directory. If not passed, it will be the account
            that created the directory, likely SYSTEM.
            Default is ``None``.

        grant_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and the basic permissions to
            grant, ie: ``{'user': {'perms': 'basic_permission'}}``. You can also
            set the ``applies_to`` setting here. The default is
            ``this_folder_subfolders_files``. Specify another ``applies_to``
            setting like this:

            .. code-block:: yaml

                {'user': {'perms': 'full_control', 'applies_to': 'this_folder'}}

            To set advanced permissions use a list for the ``perms`` parameter, ie:

            .. code-block:: yaml

                {'user': {'perms': ['read_attributes', 'read_ea'], 'applies_to': 'this_folder'}}

            Default is ``None``.

        deny_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and permissions to deny along
            with the ``applies_to`` setting. Use the same format used for the
            ``grant_perms`` parameter. Remember, deny permissions supersede
            grant permissions.
            Default is ``None``.

        inheritance (:obj:`bool`, optional):
            If ``True``, the object will inherit permissions from the parent. If
            ``False``, inheritance will be disabled. Inheritance setting will
            not apply to parent directories if they must be created.
            Default is ``True``.

        reset (:obj:`bool`, optional):
            If ``True``, the existing DACL will be cleared and replaced with the
            settings defined in this function. If ``False``, new entries will be
            appended to the existing DACL.
            Default is ``False``.

            .. versionadded:: 2018.3.0

    Returns:
        bool: ``True`` if successful

    Raises:
        OSError: Failed to create dirs and set perms

    CLI Example:

    .. code-block:: bash

        # To grant the 'Users' group 'read & execute' permissions.
        salt '*' file.makedirs_perms 'C:\\Temp\\' Administrators '{"Users": {"perms": "read_execute"}}'

        # Specify advanced attributes with a list
        salt '*' file.makedirs_perms 'C:\\Temp\\' Administrators '{"jsnuffy": {"perms": ["read_attributes", "read_ea"], "applies_to": "this_folder_files"}}'

    .. code-block:: bat

        # Locally using salt call
        salt-call file.makedirs_perms "C:\\Temp\\" Administrators "{'Users': {'perms': 'read_execute', 'applies_to': 'this_folder_only'}}"
    """
    # Expand any environment variables
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)

    # Get parent directory (head)
    head, tail = os.path.split(path)

    # If tail is empty, split head
    if not tail:
        head, tail = os.path.split(head)

    # If head and tail are defined and head is not there, recurse
    if head and tail and not os.path.exists(head):
        try:
            # Create the directory here, set inherited True because this is a
            # parent directory, the inheritance setting will only apply to the
            # target directory. Reset will be False as we only want to reset
            # the permissions on the target directory
            makedirs_perms(
                path=head,
                owner=owner,
                grant_perms=grant_perms,
                deny_perms=deny_perms,
                inheritance=True,
                reset=False,
            )
        except OSError as exc:
            # be happy if someone already created the path
            if exc.errno != errno.EEXIST:
                raise
        if tail == os.curdir:  # xxx/newdir/. exists if xxx/newdir exists
            return {}

    # Make the directory
    mkdir(
        path=path,
        owner=owner,
        grant_perms=grant_perms,
        deny_perms=deny_perms,
        inheritance=inheritance,
        reset=reset,
    )

    return True


def check_perms(
    path,
    ret=None,
    owner=None,
    grant_perms=None,
    deny_perms=None,
    inheritance=True,
    reset=False,
):
    """
    Check owner and permissions for the passed directory. This function checks
    the permissions and sets them, returning the changes made. Used by the file
    state to populate the return dict

    Args:

        path (str):
            The full path to the directory.

        ret (:obj:`dict`, optional):
            A dictionary to append changes to and return. If not passed, will
            create a new dictionary to return.
            Default is ``None``.

        owner (:obj:`str`, optional):
            The owner to set for the directory.
            Default is ``None``.

        grant_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and the basic permissions to
            check/grant, ie: ``{'user': {'perms': 'basic_permission'}}``.
            Default is ``None``.

        deny_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and permissions to
            check/deny.
            Default is ``None``.

        inheritance (:obj:`bool`, optional):
            ``True`` will check if inheritance is enabled and enable it.
            ``False`` will check if inheritance is disabled and disable it.
            Default is ``True``.

        reset (:obj:`bool`, optional):
            ``True`` will show what permissions will be removed by resetting the
            DACL. ``False`` will do nothing.
            Default is ``False``.

    Returns:
        dict: A dictionary of changes that have been made

    CLI Example:

    .. code-block:: bash

        # To see changes to C:\\Temp if the 'Users' group is given 'read & execute' permissions.
        salt '*' file.check_perms 'C:\\Temp\\' '{}' Administrators '{"Users": {"perms": "read_execute"}}'

        # Specify advanced attributes with a list
        salt '*' file.check_perms 'C:\\Temp\\' '{}' Administrators '{"jsnuffy": {"perms": ["read_attributes", "read_ea"], "applies_to": "files_only"}}'

    .. code-block:: bat

        # Locally using salt call
        salt-call file.check_perms "C:\\Temp\\" {} Administrators "{'Users': {'perms': 'read_execute', 'applies_to': 'this_folder_only'}}"

    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Path not found: {path}")

    path = os.path.expanduser(path)

    return salt.utils.win_dacl.check_perms(
        obj_name=path,
        obj_type="file",
        ret=ret,
        owner=owner,
        grant_perms=grant_perms,
        deny_perms=deny_perms,
        inheritance=inheritance,
        reset=reset,
        test_mode=__opts__["test"],
    )


def set_perms(path, grant_perms=None, deny_perms=None, inheritance=True, reset=False):
    """
    Set permissions for the given path

    Args:

        path (str):
            The full path to the directory.

        grant_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and the basic permissions to
            grant, ie: ``{'user': {'perms': 'basic_permission'}}``. You can also
            set the ``applies_to`` setting here for directories. The default for
            ``applies_to`` is ``this_folder_subfolders_files``. Specify another
            ``applies_to`` setting like this:

            .. code-block:: yaml

                {'user': {'perms': 'full_control', 'applies_to': 'this_folder'}}

            To set advanced permissions use a list for the ``perms`` parameter,
            ie:

            .. code-block:: yaml

                {'user': {'perms': ['read_attributes', 'read_ea'], 'applies_to': 'this_folder'}}

            To see a list of available attributes and applies to settings see
            the documentation for salt.utils.win_dacl.

            A value of ``None`` will make no changes to the ``grant`` portion of
            the DACL.

            Default is ``None``.

        deny_perms (:obj:`dict`, optional):
            A dictionary containing the user/group and permissions to deny along
            with the ``applies_to`` setting. Use the same format used for the
            ``grant_perms`` parameter. Remember, deny permissions supersede
            grant permissions.

            A value of ``None`` will make no changes to the ``deny`` portion of
            the DACL.

            Default is ``None``.

        inheritance (:obj:`bool`, optional):
            If ``True``, the object will inherit permissions from the parent. If
            ``False``, inheritance will be disabled. Inheritance setting will
            not apply to parent directories if they must be created.
            Default is ``True``.

        reset (:obj:`bool`, optional):
            If ``True``, the existing DCL will be cleared and replaced with the
            settings defined in this function. If ``False``, new entries will be
            appended to the existing DACL.
            Default is ``False``.

            .. versionadded:: 2018.3.0

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: If unsuccessful

    CLI Example:

    .. code-block:: bash

        # To grant the 'Users' group 'read & execute' permissions.
        salt '*' file.set_perms 'C:\\Temp\\' '{"Users": {"perms": "read_execute"}}'

        # Specify advanced attributes with a list
        salt '*' file.set_perms 'C:\\Temp\\' '{"jsnuffy": {"perms": ["read_attributes", "read_ea"], "applies_to": "this_folder_only"}}'

    .. code-block:: bat

        # Locally using salt call
        salt-call file.set_perms "C:\\Temp\\" "{'Users': {'perms': 'read_execute', 'applies_to': 'this_folder_only'}}"
    """
    return salt.utils.win_dacl.set_perms(
        obj_name=path,
        obj_type="file",
        grant_perms=grant_perms,
        deny_perms=deny_perms,
        inheritance=inheritance,
        reset=reset,
    )



"""
SmartOS grain provider

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils, salt.module.cmdmod
:platform:      SmartOS

.. versionadded:: 2017.7.0

"""

import logging
import os
import re

import salt.modules.cmdmod
import salt.utils.dictupdate
import salt.utils.json
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils

__virtualname__ = "smartos"
__salt__ = {
    "cmd.run": salt.modules.cmdmod.run,
}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load when we are on SmartOS
    """
    if salt.utils.platform.is_smartos():
        return __virtualname__
    return False


def _smartos_computenode_data():
    """
    Return useful information from a SmartOS compute node
    """
    # Provides:
    #   vms_total
    #   vms_running
    #   vms_stopped
    #   vms_type
    #   sdc_version
    #   vm_capable
    #   vm_hw_virt

    grains = {}

    # collect vm data
    vms = {}
    for vm in __salt__["cmd.run"]("vmadm list -p -o uuid,alias,state,type").split("\n"):
        vm = dict(list(zip(["uuid", "alias", "state", "type"], vm.split(":"))))
        vms[vm["uuid"]] = vm
        del vms[vm["uuid"]]["uuid"]

    # set vm grains
    grains["computenode_vms_total"] = len(vms)
    grains["computenode_vms_running"] = 0
    grains["computenode_vms_stopped"] = 0
    grains["computenode_vms_type"] = {"KVM": 0, "LX": 0, "OS": 0}
    for vm in vms:
        if vms[vm]["state"].lower() == "running":
            grains["computenode_vms_running"] += 1
        elif vms[vm]["state"].lower() == "stopped":
            grains["computenode_vms_stopped"] += 1

        if vms[vm]["type"] not in grains["computenode_vms_type"]:
            # NOTE: be prepared for when bhyve gets its own type
            grains["computenode_vms_type"][vms[vm]["type"]] = 0
        grains["computenode_vms_type"][vms[vm]["type"]] += 1

    # sysinfo derived grains
    sysinfo = salt.utils.json.loads(__salt__["cmd.run"]("sysinfo"))
    grains["computenode_sdc_version"] = sysinfo["SDC Version"]
    grains["computenode_vm_capable"] = sysinfo["VM Capable"]
    if sysinfo["VM Capable"]:
        grains["computenode_vm_hw_virt"] = sysinfo["CPU Virtualization"]

    # sysinfo derived smbios grains
    grains["manufacturer"] = sysinfo["Manufacturer"]
    grains["productname"] = sysinfo["Product"]
    grains["uuid"] = sysinfo["UUID"]

    return grains


def _smartos_zone_data():
    """
    Return useful information from a SmartOS zone
    """
    # Provides:
    #   zoneid
    #   zonename
    #   imageversion
    grains = {}

    zoneinfo = __salt__["cmd.run"]("zoneadm list -p").strip().split(":")
    grains["zoneid"] = zoneinfo[0]
    grains["zonename"] = zoneinfo[1]

    imageversion = re.compile("Image:\\s(.+)")
    grains["imageversion"] = "Unknown"
    if os.path.isfile("/etc/product"):
        with salt.utils.files.fopen("/etc/product", "r") as fp_:
            for line in fp_:
                line = salt.utils.stringutils.to_unicode(line)
                match = imageversion.match(line)
                if match:
                    grains["imageversion"] = match.group(1)

    return grains


def _smartos_zone_pkgsrc_data():
    """
    SmartOS zone pkgsrc information
    """
    # Provides:
    #   pkgsrcversion
    #   pkgsrcpath

    grains = {
        "pkgsrcversion": "Unknown",
        "pkgsrcpath": "Unknown",
    }

    # NOTE: we are specifically interested in the SmartOS pkgsrc version and path
    #       - PKG_PATH MAY be different on non-SmartOS systems, but they will not
    #         use this grains module.
    #       - A sysadmin with advanced needs COULD create a 'spin' with a totally
    #         different URL. But at that point the value would be meaning less in
    #         the context of the pkgsrcversion grain as it will not followed the
    #         SmartOS pkgsrc versioning. So 'Unknown' would be appropriate.
    pkgsrcpath = re.compile("PKG_PATH=(.+)")
    pkgsrcversion = re.compile(
        "^https?://pkgsrc.joyent.com/packages/SmartOS/(.+)/(.+)/All$"
    )
    pkg_install_paths = [
        "/opt/local/etc/pkg_install.conf",
        "/opt/tools/etc/pkg_install.conf",
    ]
    for pkg_install in pkg_install_paths:
        if os.path.isfile(pkg_install):
            with salt.utils.files.fopen(pkg_install, "r") as fp_:
                for line in fp_:
                    line = salt.utils.stringutils.to_unicode(line)
                    match_pkgsrcpath = pkgsrcpath.match(line)
                    if match_pkgsrcpath:
                        grains["pkgsrcpath"] = match_pkgsrcpath.group(1)
                        match_pkgsrcversion = pkgsrcversion.match(
                            match_pkgsrcpath.group(1)
                        )
                        if match_pkgsrcversion:
                            grains["pkgsrcversion"] = match_pkgsrcversion.group(1)
                        break

    return grains


def _smartos_zone_pkgin_data():
    """
    SmartOS zone pkgin information
    """
    # Provides:
    #   pkgin_repositories

    grains = {
        "pkgin_repositories": [],
    }

    pkginrepo = re.compile("^(?:https|http|ftp|file)://.*$")
    repositories_path = [
        "/opt/local/etc/pkgin/repositories.conf",
        "/opt/tools/etc/pkgin/repositories.conf",
    ]
    for repositories in repositories_path:
        if os.path.isfile(repositories):
            with salt.utils.files.fopen(repositories, "r") as fp_:
                for line in fp_:
                    line = salt.utils.stringutils.to_unicode(line).strip()
                    if pkginrepo.match(line):
                        grains["pkgin_repositories"].append(line)

    return grains


def smartos():
    """
    Provide grains for SmartOS
    """
    grains = {}

    if salt.utils.platform.is_smartos_zone():
        grains = salt.utils.dictupdate.update(
            grains, _smartos_zone_data(), merge_lists=True
        )
    elif salt.utils.platform.is_smartos_globalzone():
        grains = salt.utils.dictupdate.update(
            grains, _smartos_computenode_data(), merge_lists=True
        )
    grains = salt.utils.dictupdate.update(
        grains, _smartos_zone_pkgin_data(), merge_lists=True
    )
    grains = salt.utils.dictupdate.update(
        grains, _smartos_zone_pkgsrc_data(), merge_lists=True
    )

    return grains



"""
Feature flags
"""

import logging

log = logging.getLogger(__name__)


class Features:
    def __init__(self, _features=None):
        if _features is None:
            self.features = {}
        else:
            self.features = _features
        self.setup = False

    def setup_features(self, opts):
        if not self.setup:
            self.features.update(opts.get("features", {}))
        else:
            log.warning("Features already setup")

    def get(self, key, default=None):
        import salt.utils.versions

        salt.utils.versions.warn_until(
            3008,
            "Please stop checking feature flags using 'salt.features' and instead "
            "check the 'features' keyword on the configuration dictionary. The "
            "'salt.features' module will go away in {version}.",
        )
        return self.features.get(key, default)


features = Features()
setup_features = features.setup_features



"""
This module allows you to manage windows licensing via slmgr.vbs

.. code-block:: bash

    salt '*' license.install XXXXX-XXXXX-XXXXX-XXXXX-XXXXX
"""

import logging
import re

import salt.utils.platform

log = logging.getLogger(__name__)
__virtualname__ = "license"


def __virtual__():
    """
    Only work on Windows
    """
    if salt.utils.platform.is_windows():
        return __virtualname__
    return (False, "Module win_license: module only works on Windows systems.")


def installed(product_key):
    """
    Check to see if the product key is already installed.

    Note: This is not 100% accurate as we can only see the last
     5 digits of the license.

    Args:

        product_key (str): The product key to check.

    Returns:
        bool: ``True`` if the last 5 digits match the current product key,
            otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' license.installed XXXXX-XXXXX-XXXXX-XXXXX-XXXXX
    """
    cmd = r"cscript C:\Windows\System32\slmgr.vbs /dli"
    out = __salt__["cmd.run"](cmd)
    return product_key[-5:] in out


def install(product_key):
    """
    Install the given product key

    Args:

        product_key (str): The product key to install.

    Returns:
        str: The output of the ``slmgr.vbs`` script

    CLI Example:

    .. code-block:: bash

        salt '*' license.install XXXXX-XXXXX-XXXXX-XXXXX-XXXXX
    """
    cmd = rf"cscript C:\Windows\System32\slmgr.vbs /ipk {product_key}"
    return __salt__["cmd.run"](cmd)


def uninstall():
    """
    Uninstall the current product key

    CLI Example:

    .. code-block:: bash

        salt '*' license.uninstall
    """
    cmd = r"cscript C:\Windows\System32\slmgr.vbs /upk"
    return __salt__["cmd.run"](cmd)


def activate():
    """
    Attempt to activate the current machine via Windows Activation

    CLI Example:

    .. code-block:: bash

        salt '*' license.activate
    """
    cmd = r"cscript C:\Windows\System32\slmgr.vbs /ato"
    return __salt__["cmd.run"](cmd)


def licensed():
    """
    Return true if the current machine is licensed correctly

    CLI Example:

    .. code-block:: bash

        salt '*' license.licensed
    """
    cmd = r"cscript C:\Windows\System32\slmgr.vbs /dli"
    out = __salt__["cmd.run"](cmd)
    return "License Status: Licensed" in out


def info():
    """
    Return information about the license, if the license is not
    correctly activated this will return None.

    CLI Example:

    .. code-block:: bash

        salt '*' license.info
    """
    cmd = r"cscript C:\Windows\System32\slmgr.vbs /dli"
    out = __salt__["cmd.run"](cmd)

    match = re.search(
        r"Name: (.*)\r\nDescription: (.*)\r\nPartial Product Key: (.*)\r\nLicense"
        r" Status: (.*)",
        out,
        re.MULTILINE,
    )

    if match is not None:
        groups = match.groups()
        return {
            "name": groups[0],
            "description": groups[1],
            "partial_key": groups[2],
            "licensed": "Licensed" in groups[3],
        }

    return None
